import json
import logging
import typing
import warnings
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pyvista as pv
from imageio import imread, imwrite
from matplotlib import colormaps
from matplotlib.colors import ListedColormap, Normalize
from tqdm import tqdm

from geograypher.constants import NULL_TEXTURE_INT_VALUE, PATH_TYPE
from geograypher.utils.files import ensure_folder


def safe_start_xvfb():
    try:
        pv.start_xvfb()
    except OSError:
        logging.warning("Could not start xvfb because it's not supported on Windows")


def create_pv_plotter(
    off_screen: bool,
    force_xvfb: bool = False,
    plotter: typing.Union[None, pv.Plotter] = None,
):
    """Create a pyvista plotter while handling offscreen rendering

    Args:
        off_screen (bool):
            Whether the plotter should be offscreen
        force_xvfb (bool, optional):
            Should XVFB be used for rendering by default. Defaults to False.
        plotter ((None, pv.Plotter), optional):
            Existing plotter to use, will just return it if not None. Defaults to None
    """
    # If a valid plotter has not been passed in create one
    if not isinstance(plotter, pv.Plotter):
        # Catch the warning that there is not xserver running
        with warnings.catch_warnings(record=True) as w:
            # Create the plotter which may be onscreen or off
            plotter = pv.Plotter(off_screen=off_screen)

        # Start xvfb if requested or the system is not running an xserver
        if force_xvfb or (len(w) > 0 and "pyvista.start_xvfb()" in str(w[0].message)):
            # Start a headless renderer
            safe_start_xvfb()
    return plotter


def get_vis_options_from_IDs_to_labels(
    IDs_to_labels: typing.Union[None, dict],
    cmap_continous: str = "viridis",
    cmap_10_classes: str = "tab10",
    cmap_20_classes: str = "tab20",
    cmap_many_classes: str = "viridis",
):
    """Determine vis options based on a given IDs_to_labels object

    Args:
        IDs_to_labels (typing.Union[None, dict]): _description_
        cmap_continous (str, optional):
            Colormap to use if the values are continous. Defaults to "viridis".
        cmap_10_classes (str, optional):
            Colormap to use if the values are discrete and there are 10 or fewer classes. Defaults to "tab10".
        cmap_20_classes (str, optional):
            Colormap to use if the values are discrete and there are 11-20 classes. Defaults to "tab20".
        cmap_many_classes (str, optional):
            Colormap to use if there are more than 20 classes. Defaults to "viridis".

    Returns:
        dict: Containing the cmap, vmin/vmax, and whether the colormap is discrete
    """
    # This could be written in fewer lines of code but I kept it intentionally explicit

    if IDs_to_labels is None:
        # No IDs_to_labels means it's continous
        cmap = cmap_continous
        vmin = None
        vmax = None
        discrete = False
    else:
        # Otherwise, we can determine the max class ID
        max_ID = np.max(list(IDs_to_labels.keys()))

        if max_ID < 10:
            # 10 or fewer discrete classes
            cmap = cmap_10_classes
            vmin = -0.5
            vmax = 9.5
            discrete = True
        elif max_ID < 20:
            # 11-20 discrete classes
            cmap = cmap_20_classes
            vmin = -0.5
            vmax = 19.5
            discrete = True
        else:
            # More than 20 classes. There are no good discrete colormaps for this, so we generally
            # fall back on displaying it with a continous colormap
            cmap = cmap_many_classes
            vmin = None
            vmax = None
            discrete = False

    return {"cmap": cmap, "vmin": vmin, "vmax": vmax, "discrete": discrete}


def create_composite(
    RGB_image: np.ndarray,
    label_image: np.ndarray,
    label_blending_weight: float = 0.5,
    IDs_to_labels: typing.Union[None, dict] = None,
    grayscale_RGB_overlay: bool = True,
):
    """Create a three-panel composite with an RGB image and a label

    Args:
        RGB_image (np.ndarray):
            (h, w, 3) rgb image to be used directly as one panel
        label_image (np.ndarray):
            (h, w) image containing either integer labels or float scalars. Will be colormapped
            prior to display.
        label_blending_weight (float, optional):
            Opacity for the label in the blended composite. Defaults to 0.5.
        IDs_to_labels (typing.Union[None, dict], optional):
            Mapping from integer IDs to string labels. Used to compute colormap. If None, a
            continous colormap is used. Defaults to None.
        grayscale_RGB_overlay (bool):
            Convert the RGB image to grayscale in the overlay. Default is True.

    Raises:
        ValueError: If the RGB image cannot be interpreted as such

    Returns:
        np.ndarray: (h, 3*w, 3) horizontally composited image
    """
    if RGB_image.ndim != 3 or RGB_image.shape[2] != 3:
        raise ValueError("Invalid RGB error")

    if RGB_image.dtype == np.uint8:
        # Rescale to float range and implicitly cast
        RGB_image = RGB_image / 255

    vis_options = get_vis_options_from_IDs_to_labels(IDs_to_labels)
    if label_image.dtype == np.uint8 and not vis_options["discrete"]:
        # Rescale to float range and implicitly cast
        label_image = label_image / 255

    if not (label_image.ndim == 3 and label_image.shape[2] == 3):
        # If it's a one channel image make it not have a channel dim
        label_image = np.squeeze(label_image)

        cmap = plt.get_cmap(vis_options["cmap"])
        null_mask = np.logical_or(
            label_image == NULL_TEXTURE_INT_VALUE,
            np.logical_not(np.isfinite(label_image)),
        )
        if not vis_options["discrete"]:
            # Shift
            label_image = label_image - np.nanmin(label_image)
            # Find the max value that's not the null vlaue
            valid_pixels = label_image[np.logical_not(null_mask)]
            if valid_pixels.size > 0:
                # TODO this might have to be changed to nanmax in the future
                max_value = np.max(valid_pixels)
                # Scale
                label_image = label_image / max_value

        # Perform the colormapping
        label_image = cmap(label_image)[..., :3]
        # Mask invalid values
        label_image[null_mask] = 0

    # Create a blended image
    if grayscale_RGB_overlay:
        RGB_for_composite = np.tile(
            np.mean(RGB_image, axis=2, keepdims=True), (1, 1, 3)
        )
    else:
        RGB_for_composite = RGB_image
    overlay = ((1 - label_blending_weight) * RGB_for_composite) + (
        label_blending_weight * label_image
    )
    # Concatenate the images horizonally
    composite = np.concatenate((label_image, RGB_image, overlay), axis=1)
    # Cast to np.uint8 for saving
    composite = (composite * 255).astype(np.uint8)
    return composite


def read_img_npy(filename):
    try:
        return imread(filename)
    except:
        pass

    try:
        return np.load(filename)
    except:
        pass


def show_segmentation_labels(
    label_folder: PATH_TYPE,
    image_folder: PATH_TYPE,
    savefolder: typing.Union[None, PATH_TYPE] = None,
    num_show: int = 10,
    label_suffix: str = ".png",
    image_suffix: str = ".JPG",
    IDs_to_labels: typing.Optional[dict] = None,
) -> None:
    """
    Visualize and optionally save composite images showing segmentation labels overlaid
    on their corresponding images.

    Args:
        label_folder (PATH_TYPE): Path to the folder containing label images.
        image_folder (PATH_TYPE): Path to the folder containing original images.
        savefolder (PATH_TYPE, optional): If provided, composites are saved here;
            otherwise, they are displayed using pyplot. Defaults to None.
        num_show (int): Number of samples to show or save. Defaults to 10.
        label_suffix (str): Suffix for label image files. Defaults to ".png".
        image_suffix (str): Suffix for image files. Defaults to ".JPG".
        IDs_to_labels (dict, optional): Mapping from label IDs to class names.
            If None, will attempt to load from label_folder. Defaults to None.
    """
    # Find all label files in the label_folder and shuffle them
    rendered_files = list(Path(label_folder).rglob("*" + label_suffix))
    np.random.shuffle(rendered_files)

    # Ensure the save folder exists if saving output
    if savefolder is not None:
        ensure_folder(savefolder)

    # Attempt to load IDs_to_labels from a JSON file if not provided
    if (
        IDs_to_labels is None
        and (IDs_to_labels_file := Path(label_folder, "IDs_to_labels.json")).exists()
    ):
        with open(IDs_to_labels_file, "r") as infile:
            IDs_to_labels = json.load(infile)
            IDs_to_labels = {int(k): v for k, v in IDs_to_labels.items()}
    # Iterate through a subset of label files and create composites
    for i, rendered_file in tqdm(
        enumerate(rendered_files[:num_show]),
        desc="Showing segmentation labels",
        total=min(len(rendered_files), num_show),
    ):
        # Find the corresponding image file by matching the label path to an image path.
        # Assumes that image_folder and label_folder have the same subdirectory structure
        # and file stems for corresponding images/labels.
        image_file = Path(
            image_folder, rendered_file.relative_to(label_folder)
        ).with_suffix(image_suffix)

        # Read the image and label data
        image = imread(image_file)
        render = read_img_npy(rendered_file)
        # Create a composite visualization
        composite = create_composite(image, render, IDs_to_labels=IDs_to_labels)

        if savefolder is None:
            # Display the composite
            plt.imshow(composite)
            plt.show()
        else:
            # Save the composite to the output folder
            output_file = Path(savefolder, f"rendered_label_{i:03}.png")
            imwrite(output_file, composite)


def visualize_intersections_in_pyvista(
    plotter: pv.Plotter,
    ray_starts: np.ndarray,
    ray_ends: np.ndarray,
    community_IDs: np.ndarray,
    community_points: np.ndarray,
) -> None:
    """
    Visualize the given grouped rays and detected points in a pyvista plotter.

    Arguments:
        plotter (pv.Plotter):
            Existing pyvista plotter to add intersection lines/points to
        ray_starts ((N, 3) np.ndarray):
            The 3D locations of the starting points of N rays
        ray_ends ((N, 3) np.ndarray):
            The 3D locations of the ending points of N rays
        community_IDs ((N,) np.ndarray):
            The IDs for groups of rays. For example, if there are 10 rays with 5 in group 0
            and 5 in group 1, this would be [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]
        community_points ((M, 3) np.ndarray):
            One 3D point per community, indicating the center of the grouped rays. In the
            example above, this would be an (2, 3) array with 2 points.
    """

    # Interweave the points as line_segments_from_points desires
    interwoven = np.empty((2 * len(ray_starts), 3), dtype=ray_starts.dtype)
    interwoven[0::2] = ray_starts
    interwoven[1::2] = ray_ends

    # Show the line segments
    lines_mesh = pv.line_segments_from_points(interwoven)
    plotter.add_mesh(
        lines_mesh,
        scalars=community_IDs,
        label="Rays, colored by community ID",
    )

    # Show the triangulated communtities as red spheres
    detected_points = pv.PolyData(community_points)
    plotter.add_points(
        detected_points,
        color="r",
        render_points_as_spheres=True,
        point_size=10,
        label="Triangulated locations",
    )
    plotter.add_legend()


def visualize_intersections_as_mesh(
    ray_starts: np.ndarray,
    ray_ends: np.ndarray,
    community_IDs: np.ndarray,
    community_points: np.ndarray,
    out_dir: PATH_TYPE,
    batch: int = 250,
    cube_side_len: float = 0.2,
) -> None:
    """
    Arguments:
        ray_starts ((N, 3) np.ndarray):
            The 3D locations of the starting points of N rays
        ray_ends ((N, 3) np.ndarray):
            The 3D locations of the ending points of N rays
        community_IDs ((N,) np.ndarray):
            The IDs for groups of rays. For example, if there are 10 rays with 5 in group 0
            and 5 in group 1, this would be [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]
        community_points ((M, 3) np.ndarray):
            One 3D point per community, indicating the center of the grouped rays. In the
            example above, this would be an (2, 3) array with 2 points.
        out_dir (PATH_TYPE):
        batch (int):
            Defaults to 250.
        cube_side_len (float):
            Defaults to 0.2

    Saves:
        out_dir / rays.ply
        out_dir / points.ply
    """

    # Short-circuit on empty
    if len(ray_starts) == 0:
        return

    # Enforce Path type
    out_dir = Path(out_dir)

    # Split the communities into an (unforunately limited) color set
    norm = Normalize(vmin=np.nanmin(community_IDs), vmax=np.nanmax(community_IDs))
    cmap = colormaps["tab20"]

    # Build up cylinders in batches
    n_batches = int(np.ceil(len(community_IDs) / batch))
    cylinder_polydata = None
    for i in tqdm(range(n_batches), desc="Building cylinders"):
        islice = slice(i * batch, min((i + 1) * batch, len(community_IDs)))
        batched = merge_cylinders(
            starts=ray_starts[islice],
            ends=ray_ends[islice],
            community_IDs=community_IDs[islice],
            cmap=cmap,
            norm=norm,
        )
        # Merge with previous cylinders
        if cylinder_polydata is None:
            cylinder_polydata = batched
        else:
            cylinder_polydata = cylinder_polydata.merge(batched)

    if cylinder_polydata is not None:
        path = out_dir / "rays.ply"
        print(f"Saving visualized cylinders to {path}")
        cylinder_polydata.save(path, texture="RGB")

    # The cube merging is much less costly than the cylinders, and thus far hasn't
    # required batching
    cube_polydata = None
    for community_ID, point in enumerate(
        tqdm(community_points, desc="Building points")
    ):
        # Build a cube and set the color
        cube = pv.Cube(
            center=point,
            x_length=cube_side_len,
            y_length=cube_side_len,
            z_length=cube_side_len,
        )
        color = (np.array(cmap(norm(community_ID)))[:3] * 255).astype(np.uint8)
        cube.point_data["RGB"] = np.tile(color, (cube.n_points, 1))
        # Merge with previous cubes
        if cube_polydata is None:
            cube_polydata = cube
        else:
            cube_polydata = cube_polydata.merge(cube)

    if cube_polydata is not None:
        path = out_dir / "points.ply"
        print(f"Saving visualized cubes to {path}")
        cube_polydata.save(path, texture="RGB")


def merge_cylinders(
    starts: np.ndarray,
    ends: np.ndarray,
    community_IDs: np.ndarray,
    cmap: ListedColormap,
    norm: Normalize,
) -> pv.PolyData:
    """
    Create and merge a set of cylinders one by one.

    Arguments:
        ray_starts ((N, 3) np.ndarray):
            The 3D locations of the starting points of N rays
        ray_ends ((N, 3) np.ndarray):
            The 3D locations of the ending points of N rays
        community_IDs ((N,) np.ndarray):
            The IDs for groups of rays. For example, if there are 10 rays with 5 in group 0
            and 5 in group 1, this would be [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]
        cmap (matplotlib.colors.ListedColormap):
            Colormap to use on the normalized community IDs
        norm (matplotlib.colors.Normalize):
            Normalization function that spans from min to max of the community IDs

    Returns: pv.Polydata mesh representing the given rays with cylinders.
    """

    polydata = None
    for start, end, community_ID in zip(starts, ends, community_IDs):

        # Build a cylinder
        center = (start + end) / 2
        direction = end - start
        height = np.linalg.norm(direction)
        if height == 0:
            continue
        direction = direction / height
        # Some of the chosen parameters (low resolution, no capping) are to reduce
        # polygon faces when dealing with large numbers of cylinders
        cyl = pv.Cylinder(
            center=center,
            direction=direction,
            radius=0.05,
            height=height,
            resolution=3,
            capping=False,
        )

        # Color the cylinder
        color = (np.array(cmap(norm(community_ID)))[:3] * 255).astype(np.uint8)
        cyl["scalars"] = np.full(cyl.n_points, community_ID)
        cyl.point_data["RGB"] = np.tile(color, (cyl.n_points, 1))

        # And merge it into the scene
        if polydata is None:
            polydata = cyl
        else:
            polydata = polydata.merge(cyl)

    return polydata



import argparse
import typing
from pathlib import Path

import fiona
import geopandas as gpd
import numpy as np
import pyproj
import shapely

from geograypher.cameras import MetashapeCameraSet
from geograypher.constants import PATH_TYPE
from geograypher.meshes import TexturedPhotogrammetryMesh
from geograypher.meshes.derived_meshes import TexturedPhotogrammetryMeshChunked
from geograypher.utils.visualization import show_segmentation_labels


# TODO Consider adding a mesh screenshot option or a vis savefolder
def render_labels(
    mesh_file: PATH_TYPE,
    cameras_file: PATH_TYPE,
    image_folder: PATH_TYPE,
    texture: typing.Union[PATH_TYPE, np.ndarray, None],
    render_savefolder: PATH_TYPE,
    mesh_CRS: pyproj.CRS,
    original_image_folder: typing.Union[PATH_TYPE, None] = None,
    subset_images_savefolder: typing.Union[PATH_TYPE, None] = None,
    texture_column_name: typing.Union[str, None] = None,
    DTM_file: typing.Union[PATH_TYPE, None] = None,
    ground_height_threshold: typing.Union[float, None] = None,
    render_ground_class: bool = False,
    textured_mesh_savefile: typing.Union[PATH_TYPE, None] = None,
    ROI: typing.Union[PATH_TYPE, gpd.GeoDataFrame, shapely.MultiPolygon, None] = None,
    mesh_ROI_buffer_radius_meters: float = 50,
    cameras_ROI_buffer_radius_meters: float = 150,
    IDs_to_labels: typing.Union[dict, None] = None,
    render_image_scale: float = 1,
    mesh_downsample: float = 1,
    n_render_clusters: typing.Union[int, None] = None,
    cast_to_uint8: bool = True,
    save_as_npy: bool = False,
    vis: bool = False,
    mesh_vis_file: typing.Union[PATH_TYPE, None] = None,
    labels_vis_folder: typing.Union[PATH_TYPE, None] = None,
):
    """Renders image-based labels using geospatial ground truth data

    Args:
        mesh_file (PATH_TYPE):
            Path to the Metashape-exported mesh file
        cameras_file (PATH_TYPE):
            Path to the MetaShape-exported .xml cameras file
        image_folder (PATH_TYPE):
            Path to the folder of images used to create the mesh
        texture (typing.Union[PATH_TYPE, np.ndarray, None]):
            See TexturedPhotogrammetryMesh.load_texture
        render_savefolder (PATH_TYPE):
            Where to save the rendered labels
        mesh_CRS: (pyproj.CRS):
            The vertex coordinates of the input mesh should be interpreteted in this coordinate
            references system to georeference them. Since meshes are not commonly used for
            geospatial tasks, there isn't a common standard for encoding this information in the mesh.
        original_image_folder (typing.Union[PATH_TYPE, None], optional):
            Where the images were when photogrammetry was run. Metashape saves imagenames with an
            absolute path which can cause issues. If this argument is provided, this path is removed
            from the start of each image file name, which allows the camera set to be used with a
            moved folder of images specified by `image_folder`. Defaults to None.
        subset_images_savefolder (typing.Union[PATH_TYPE, None], optional):
            Where to save the subset of images for which labels are generated. Defaults to None.
        texture_column_name (typing.Union[str, None], optional):
            Column to use in vector file for texture information". Defaults to None.
        DTM_file (typing.Union[PATH_TYPE, None], optional):
            Path to a DTM file to use for ground thresholding. Defaults to None.
        ground_height_threshold (typing.Union[float, None], optional):
            Set points under this height to ground. Only applicable if DTM_file is provided. Defaults to None.
        render_ground_class (bool, optional):
            Should the ground class be included in the renders or deleted.. Defaults to False.
        textured_mesh_savefile (typing.Union[PATH_TYPE, None], optional):
            Where to save the textured and subsetted mesh, if needed in the future. Defaults to None.
        ROI (typing.Union[PATH_TYPE, gpd.GeoDataFrame, shapely.MultiPolygon, None], optional):
            The region of interest to render labels for. Defaults to None.
        mesh_ROI_buffer_radius_meters (float, optional):
            The distance in meters to include around the ROI for the mesh. Defaults to 50.
        cameras_ROI_buffer_radius_meters (float, optional):
            The distance in meters to include around the ROI for the cameras. Defaults to 150.
        IDs_to_labels (typing.Union[None, dict], optional):
            Mapping between the integer labels and string values for the classes. Defaults to None.
        render_image_scale (float, optional):
            Downsample the images to this fraction of the size for increased performance but lower quality. Defaults to 1.
        mesh_downsample (float, optional):
            Downsample the mesh to this fraction of vertices for increased performance but lower quality. Defaults to 1.
        n_render_clusters (typing.Union[int, None]):
            If set, break the camera set and mesh into this many clusters before rendering. This is
            useful for large meshes that are otherwise very slow. Defaults to None.
        cast_to_uint8 (bool, optional):
            If True, cast the rendered labels to uint8. If False, it is cast to uint16 or uint32, if not saved as .npy.
            Defaults to True.
        save_as_npy (bool, optional):
            If True, save the rendered labels are float64 and saved as numpy arrays. Defaults to False.
        mesh_vis (typing.Union[PATH_TYPE, None])
            Path to save the visualized mesh instead of showing it interactively. Only applicable if vis=True. Defaults to None.
        labels_vis (typing.Union[PATH_TYPE, None])
            Defaults to None.
    """
    ## Determine the ROI
    # If the ROI is unset and the texture is a geodataframe, set the ROI to that
    if ROI is None and isinstance(texture, gpd.GeoDataFrame):
        ROI = texture
    elif ROI is None and isinstance(texture, (str, Path)):
        try:
            ROI = gpd.read_file(texture)
        except fiona.errors.DriverError:
            pass

    ## Create the camera set
    # This is done first because it's often faster than mesh operations which
    # makes it a good place to check for failures
    camera_set = MetashapeCameraSet(
        cameras_file, image_folder, original_image_folder=original_image_folder
    )

    if ROI is not None:
        # Extract cameras near the training data
        camera_set = camera_set.get_subset_ROI(
            ROI=ROI, buffer_radius=cameras_ROI_buffer_radius_meters, is_geospatial=True
        )
    # If requested, save out the images corresponding to this subset of cameras.
    # This is useful for model training.
    if subset_images_savefolder is not None:
        camera_set.save_images(subset_images_savefolder)

    # Select whether to use a class that renders by chunks or not
    MeshClass = (
        TexturedPhotogrammetryMesh
        if n_render_clusters is None
        else TexturedPhotogrammetryMeshChunked
    )

    ## Create the textured mesh
    mesh = MeshClass(
        mesh_file,
        input_CRS=mesh_CRS,
        downsample_target=mesh_downsample,
        texture=texture,
        texture_column_name=texture_column_name,
        ROI=ROI,
        ROI_buffer_meters=mesh_ROI_buffer_radius_meters,
        IDs_to_labels=IDs_to_labels,
    )

    ## Set the ground class if applicable
    if DTM_file is not None and ground_height_threshold is not None:
        # The ground ID will be set to the next value if None, or np.nan if np.nan
        ground_ID = None if render_ground_class else np.nan
        mesh.label_ground_class(
            DTM_file=DTM_file,
            height_above_ground_threshold=ground_height_threshold,
            only_label_existing_labels=True,
            ground_class_name="GROUND",
            ground_ID=ground_ID,
            set_mesh_texture=True,
        )

    # Save the textured and subsetted mesh, if applicable
    if textured_mesh_savefile is not None:
        mesh.save_mesh(textured_mesh_savefile)

    # Show the cameras and mesh if requested
    if vis or mesh_vis_file is not None:
        mesh.vis(camera_set=camera_set, screenshot_filename=mesh_vis_file)

    # Include n_render_clusters as an optional keyword argument, if provided. This is only applicable
    # if this mesh is a TexturedPhotogrammetryMeshChunked object
    render_kwargs = (
        {} if n_render_clusters is None else {"n_clusters": n_render_clusters}
    )
    # Render the labels and save them. This is the slow step.
    mesh.save_renders(
        camera_set=camera_set,
        render_image_scale=render_image_scale,
        save_native_resolution=True,
        output_folder=render_savefolder,
        make_composites=False,
        cast_to_uint8=cast_to_uint8,
        save_as_npy=save_as_npy,
        **render_kwargs,
    )

    if vis or labels_vis_folder is not None:
        # Show some examples of the rendered labels side-by-side with the real images
        show_segmentation_labels(
            label_folder=render_savefolder,
            image_folder=image_folder,
            savefolder=labels_vis_folder,
            label_suffix=".tif",
            num_show=10,
        )


def parse_args():
    """Parse and return arguements

    Returns:
        argparse.Namespace: Arguments
    """
    description = (
        "This script renders labels onto individual images using geospatial textures. "
        + "By default is uses the example data. All arguments are passed to "
        + "geograypher.entrypoints.workflow_functions.render_labels "
        + "which has the following documentation:\n\n"
        + render_labels.__doc__
    )
    # Ideally we'd include the defaults for each argument, but there is no help text so the
    # ArgumentDefaultsHelpFormatter formatter doesn't show them
    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.RawDescriptionHelpFormatter
    )

    # Add arguments
    parser.add_argument("--mesh-file", type=Path, required=True)
    parser.add_argument("--mesh-CRS", required=True)
    parser.add_argument("--cameras-file", type=Path, required=True)
    parser.add_argument("--image-folder", type=Path, required=True)
    parser.add_argument("--texture", type=Path, required=True)
    parser.add_argument("--render-savefolder", type=Path, required=True)
    parser.add_argument("--original-image-folder", type=Path)
    parser.add_argument("--subset-images-savefolder", type=Path)
    parser.add_argument("--texture-column-name")
    parser.add_argument("--DTM-file")
    parser.add_argument("--ground-height-threshold", type=float, default=2.0)
    parser.add_argument("--render-ground-class", action="store_true")
    parser.add_argument("--textured-mesh-savefile")
    parser.add_argument("--ROI")
    parser.add_argument("--mesh-ROI-buffer-radius-meters", default=50, type=float)
    parser.add_argument("--cameras-ROI-buffer-radius-meters", default=100, type=float)
    parser.add_argument("--render-image-scale", type=float, default=1)
    parser.add_argument("--mesh-downsample", type=float, default=1)
    parser.add_argument("--vis", action="store_true")
    parser.add_argument("--mesh-vis-file", type=Path)
    parser.add_argument("--labels-vis-folder", type=Path)

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    # Parse command line args
    args = parse_args()
    # Pass all the arguments command line options to render_labels
    render_labels(**args.__dict__)



import argparse
import json
import math
import typing
from pathlib import Path

import numpy as np
import pyproj

from geograypher.cameras import MetashapeCameraSet, SegmentorPhotogrammetryCameraSet
from geograypher.constants import EXAMPLE_IDS_TO_LABELS, PATH_TYPE
from geograypher.meshes import TexturedPhotogrammetryMesh
from geograypher.meshes.derived_meshes import TexturedPhotogrammetryMeshChunked
from geograypher.predictors.derived_segmentors import LookUpSegmentor
from geograypher.utils.files import ensure_containing_folder
from geograypher.utils.indexing import find_argmax_nonzero_value


def aggregate_images(
    mesh_file: PATH_TYPE,
    cameras_file: PATH_TYPE,
    image_folder: PATH_TYPE,
    label_folder: PATH_TYPE,
    mesh_CRS: pyproj.CRS,
    original_image_folder: typing.Union[PATH_TYPE, None] = None,
    subset_images_folder: typing.Union[PATH_TYPE, None] = None,
    filename_regex: typing.Optional[str] = None,
    take_every_nth_camera: typing.Union[int, None] = 100,
    DTM_file: typing.Union[PATH_TYPE, None] = None,
    height_above_ground_threshold: float = 2.0,
    ROI: typing.Union[PATH_TYPE, None] = None,
    ROI_buffer_radius_meters: float = 50,
    IDs_to_labels: typing.Union[dict, str, None] = None,
    mesh_downsample: float = 1.0,
    n_aggregation_clusters: typing.Union[int, None] = None,
    n_cameras_per_aggregation_cluster: typing.Union[int, None] = None,
    aggregate_image_scale: float = 1.0,
    aggregated_face_values_savefile: typing.Union[PATH_TYPE, None] = None,
    predicted_face_classes_savefile: typing.Union[PATH_TYPE, None] = None,
    top_down_vector_projection_savefile: typing.Union[PATH_TYPE, None] = None,
    vis: bool = False,
):
    """Aggregate labels from multiple viewpoints onto the surface of the mesh

    Args:
        mesh_file (PATH_TYPE):
            Path to the Metashape-exported mesh file
        cameras_file (PATH_TYPE):
            Path to the MetaShape-exported .xml cameras file
        image_folder (PATH_TYPE):
            Path to the folder of images used to create the mesh
        filename_regex (str, optional):
            Use only images with paths matching this regex
        label_folder (PATH_TYPE):
            Path to the folder of labels to be aggregated onto the mesh. Must be in the same
            structure as the images
        mesh_CRS (pyproj.CRS):
            The CRS to interpret the mesh in.
        original_image_folder (typing.Union[PATH_TYPE, None], optional):
            Where the images were when photogrammetry was run. Metashape saves imagenames with an
            absolute path which can cause issues. If this argument is provided, this path is removed
            from the start of each image file name, which allows the camera set to be used with a
            moved folder of images specified by `image_folder`. Defaults to None.
        subset_images_folder (typing.Union[PATH_TYPE, None], optional):
            Use only images from this subset. Defaults to None.
        take_every_nth_camera (typing.Union[int, None], optional):
            Downsample the camera set to only every nth camera if set. Defaults to None.
        DTM_file (typing.Union[PATH_TYPE, None], optional):
            Path to a digital terrain model file to remove ground points. Defaults to None.
        height_above_ground_threshold (float, optional):
            Height in meters above the DTM to consider ground. Only used if DTM_file is set.
            Defaults to 2.0.
        ROI (typing.Union[PATH_TYPE, None], optional):
            Geofile region of interest to crop the mesh to. Defaults to None.
        ROI_buffer_radius_meters (float, optional):
            Keep points within this distance of the provided ROI object, if unset, everything will
            be kept. Defaults to 50.
        IDs_to_labels (typing.Union[dict, None], optional):
            Maps from integer IDs to human-readable class name labels. Defaults to None.
        mesh_downsample (float, optional):
            Downsample the mesh to this fraction of vertices for increased performance but lower
            quality. Defaults to 1.0.
        n_aggregation_clusters (typing.Union[int, None]):
            If set, aggregate with this many clusters. Defaults to None.
        n_cameras_per_aggregation_cluster (typing.Union[int, None]):
            If set, and n_aggregation_clusters is not, use to compute a number of clusters such that
            each cluster has this many cameras. Defaults to None.
        aggregate_image_scale (float, optional):
            Downsample the labels before aggregation for faster runtime but lower quality. Defaults
            to 1.0.
        aggregated_face_values_savefile (typing.Union[PATH_TYPE, None], optional):
            Where to save the aggregated image values as a numpy array. Defaults to None.
        predicted_face_classes_savefile (typing.Union[PATH_TYPE, None], optional):
            Where to save the most common label per face texture as a numpy array. Defaults to None.
        top_down_vector_projection_savefile (typing.Union[PATH_TYPE, None], optional):
            Where to export the predicted map. Defaults to None.
        vis (bool, optional):
            Show the mesh model and predicted results. Defaults to False.
    """

    if isinstance(IDs_to_labels, str):
        IDs_to_labels = {
            int(k): v for k, v in json.load(open(IDs_to_labels, "r")).items()
        }

    ## Create the camera set
    # Do the camera operations first because they are fast and good initial error checking
    camera_set = MetashapeCameraSet(
        cameras_file,
        image_folder,
        original_image_folder=original_image_folder,
        validate_images=True,
    )

    # If the ROI is not None, subset to cameras within a buffer distance of the ROI
    # TODO let get_subset_ROI accept a None ROI and return the full camera set
    if subset_images_folder is not None:
        camera_set = camera_set.get_cameras_in_folder(subset_images_folder)

    # Subset based on regex if requested
    if filename_regex is not None:
        camera_set = camera_set.get_cameras_matching_filename_regex(
            filename_regex=filename_regex
        )

    # If you only want to take every nth camera, helpful for initial testing
    if take_every_nth_camera is not None:
        camera_set = camera_set.get_subset_cameras(
            range(0, len(camera_set), take_every_nth_camera)
        )

    if ROI is not None and ROI_buffer_radius_meters is not None:
        # Extract cameras near the training data
        camera_set = camera_set.get_subset_ROI(
            ROI=ROI, buffer_radius=ROI_buffer_radius_meters
        )

    # If the number of aggregation clusters is not set but the number of cameras per cluster is,
    # then compute it
    if n_aggregation_clusters is None and n_cameras_per_aggregation_cluster is not None:
        n_aggregation_clusters = int(
            math.ceil(len(camera_set) / n_cameras_per_aggregation_cluster)
        )

    # Choose whether to use a mesh class that aggregates by clusters of cameras and chunks of the mesh
    MeshClass = (
        TexturedPhotogrammetryMesh
        if n_aggregation_clusters is None
        else TexturedPhotogrammetryMeshChunked
    )
    ## Create the mesh
    mesh = MeshClass(
        mesh_file,
        input_CRS=mesh_CRS,
        ROI=ROI,
        ROI_buffer_meters=ROI_buffer_radius_meters,
        IDs_to_labels=IDs_to_labels,
        downsample_target=mesh_downsample,
    )

    # Show the mesh if requested
    if vis:
        mesh.vis(camera_set=camera_set)

    # Create a segmentor object to load in the predictions
    segmentor = LookUpSegmentor(
        base_folder=image_folder,
        lookup_folder=label_folder,
        num_classes=np.max(list(mesh.get_IDs_to_labels().keys())) + 1,
    )
    # Create a camera set that returns the segmented images instead of the original ones
    segmentor_camera_set = SegmentorPhotogrammetryCameraSet(
        camera_set, segmentor=segmentor
    )

    # Create the potentially-empty dict of kwargs to match what this class expects
    n_clusters_kwargs = (
        {} if n_aggregation_clusters is None else {"n_clusters": n_aggregation_clusters}
    )

    ## Perform aggregation, this is the slow step
    aggregated_face_labels, _ = mesh.aggregate_projected_images(
        segmentor_camera_set,
        aggregate_img_scale=aggregate_image_scale,
        **n_clusters_kwargs,
    )

    # If requested, save this data
    if aggregated_face_values_savefile is not None:
        ensure_containing_folder(aggregated_face_values_savefile)
        np.save(aggregated_face_values_savefile, aggregated_face_labels)

    # Find the index of the most common class per face, with faces with no predictions set to nan
    predicted_face_classes = find_argmax_nonzero_value(
        aggregated_face_labels, keepdims=True
    )

    # If requested, label the ground faces
    if DTM_file is not None and height_above_ground_threshold is not None:
        predicted_face_classes = mesh.label_ground_class(
            labels=predicted_face_classes,
            height_above_ground_threshold=height_above_ground_threshold,
            DTM_file=DTM_file,
            ground_ID=np.nan,
            set_mesh_texture=False,
        )

    if predicted_face_classes_savefile is not None:
        ensure_containing_folder(predicted_face_classes_savefile)
        np.save(predicted_face_classes_savefile, predicted_face_classes)

    if vis:
        # Show the mesh with predicted classes
        mesh.vis(vis_scalars=predicted_face_classes)

    # If the vector file should be exported
    if top_down_vector_projection_savefile is not None:
        # Compute the label names
        if IDs_to_labels is not None:
            # This ensures that any missing keys are replaced with None so proper indexing is retained
            label_names = [
                IDs_to_labels.get(i, None)
                for i in range(max(list(IDs_to_labels.keys())) + 1)
            ]
        else:
            label_names = None
        # Export the 2D top down projection
        mesh.export_face_labels_vector(
            face_labels=np.squeeze(predicted_face_classes),
            export_file=top_down_vector_projection_savefile,
            vis=vis,
            label_names=label_names,
        )


def parse_args():
    description = (
        "This script aggregates predictions from individual images onto the mesh. This aggregated "
        + "prediction can then be exported into geospatial coordinates. The default option is to "
        + "use the provided example data. All of the arguments are passed to "
        + "geograypher.entrypoints.workflow_functions.aggregate_images "
        + "which has the following documentation:\n\n"
        + aggregate_images.__doc__
    )
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter, description=description
    )
    parser.add_argument("--mesh-file", required=True)
    parser.add_argument("--cameras-file", required=True)
    parser.add_argument("--image-folder", required=True)
    parser.add_argument("--label-folder", required=True)
    parser.add_argument("--mesh-CRS", required=True)
    parser.add_argument("--original-image-folder", type=Path)
    parser.add_argument("--subset-images-folder", type=Path)
    parser.add_argument("--take-every-nth-camera", type=int)
    parser.add_argument("--DTM-file", type=Path)
    parser.add_argument("--height-above-ground-threshold", type=float, default=2)
    parser.add_argument("--ROI")
    parser.add_argument("--ROI-buffer-radius-meters", default=50, type=float)
    parser.add_argument("--IDs-to-labels", default=EXAMPLE_IDS_TO_LABELS)
    parser.add_argument("--mesh-downsample", type=float, default=1.0)
    parser.add_argument("--aggregate-image-scale", type=float, default=0.25)
    parser.add_argument("--n-aggregation-clusters", type=int)
    parser.add_argument("--aggregated-face-values-savefile", type=Path)
    parser.add_argument("--predicted-face-classes-savefile", type=Path)
    parser.add_argument(
        "--top-down-vector-projection-savefile", default="vis/predicted_map.geojson"
    )
    parser.add_argument("--vis", action="store_true")

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    # Parse command line args
    args = parse_args()
    # Pass command line args to aggregate_images
    aggregate_images(**args.__dict__)



from itertools import product
from typing import List, Tuple

import numpy as np
import pyvista as pv

from geograypher.cameras.cameras import PhotogrammetryCamera, PhotogrammetryCameraSet


def make_simple_camera_set() -> PhotogrammetryCameraSet:
    """
    Create a simple camera looking down at the origin from above, spaced to line up
    perfectly with the simple mesh.
    """

    # Focal length in pixels
    focal = 100
    # Sensor width/height in pixels
    sensor = 200

    return PhotogrammetryCameraSet(
        cameras=[
            PhotogrammetryCamera(
                image_filename=None,
                cam_to_world_transform=downward_view(
                    scene_width=4,  # Known width of the simple mesh in X/Y
                    focal=focal,
                    sensor_width=sensor,
                ),
                f=focal,
                cx=0,
                cy=0,
                image_width=sensor,
                image_height=sensor,
                local_to_epsg_4978_transform=np.eye(4),
            )
        ],
        local_to_epsg_4978_transform=np.eye(4),
    )


def downward_view(scene_width, focal, sensor_width):
    """
    The equation is just a ratio of triangles
    Camera distance (m) = Scene width (m) * focal length (px) / sensor width (px)

       |.
       |   .               .|
    ↑  |       .        .   | ↑
    sw |  ←cd→   [cam]  ←f→ | sensor w
    ↓  |       .        .   | ↓
       |   .               .|
       |.

    Note that the orientation is set so the camera is looking down. If you imagine
    rotating a right-hand reference frame by 180 so Z (out of the camera) points
    down, you end up with Y pointing along -Y and Z pointing along -Z.
    """
    return np.array(
        [
            [1, 0, 0, 0],
            [0, -1, 0, 0],
            [0, 0, -1, scene_width * focal / sensor_width],
            [0, 0, 0, 1],
        ]
    )


def make_simple_mesh(
    pixels: np.ndarray, color: List[int], background: int = 50, buffer: int = 0
) -> Tuple[pv.PolyData, np.ndarray]:
    """
    Create a flat mesh with the given pixels colored. Designed to line up
    with the simple camera so that 1 interval between points = 1 pixel.

    Arguments:
        pixels (ndarray): (Q, 2) array of (i, j) pixel values we want to color
        color (List[int]): 3-element 0-255 color to set the pixels to
        background (int):
        buffer (int): Number of pixels to offset around each pixel with the
            same color

    Returns: Two element tuple of
        [0] pyvista mesh of the plane
        [1] (M, 3) color vector for each point in the mesh
    """

    # Define the number of pixels we want to support. 200 intervals (pixels) =
    # 201 points in the mesh.
    N = 201

    plane = pv.Plane(
        center=(0, 0, 0),
        direction=(0, 0, 1),  # normal pointing up
        i_size=4,  # From -2 to 2
        j_size=4,  # From -2 to 2
        i_resolution=N - 1,
        j_resolution=N - 1,
    )
    if plane is None:
        raise ValueError("Failed to create plane mesh")

    # Triangulate to ensure we have triangular faces
    plane = plane.triangulate()
    if plane is None:
        raise ValueError("Failed to create plane triangulation")

    # Create point colors
    n_points = plane.n_points
    point_colors = np.full((n_points, 3), fill_value=background, dtype=np.uint8)

    # Fill them in pixel by pixel
    for pixel in pixels:
        pixel_idx(point_colors, *pixel, stride=N, color=color, buffer=buffer)

    ##########################################################################
    # For visual debugging purposes, uncomment this code, it will color the
    # edges of the mesh in bright colors.
    # point_colors[0] = [255, 0, 0]
    # point_colors[1] = [0, 0, 255]
    # point_colors[:2*N] = [255, 0, 0]
    # point_colors[-2*N:] = [0, 255, 0]
    # point_colors[0::N] = [0, 0, 255]
    # point_colors[1::N] = [0, 0, 255]
    # point_colors[N-2::N] = [0, 255, 255]
    # point_colors[N-1::N] = [0, 255, 255]
    ##########################################################################

    return plane, point_colors


def pixel_idx(
    vector: np.ndarray, i: int, j: int, stride: int, color: List[int], buffer: int = 0
) -> None:
    """
    Helper to turn unflattened ij pixels into the color vector around
    that mesh area. Note that we have to invert the i dimension b/c of
    how plane texturing is applied vs. how images are rendered. This
    way (i, j) will correspond to (i, j) in the image.

    Arguments:
        vector (ndarray): Shape (M*N,) array, flattened (M, N) image
        i, j (int): row, column location we want to set a color
        stride: N, a.k.a. the number of columns in the unflattened image
        color (List[int]): 3-element 0-255 color to set the pixel to
        buffet (int): Number of pixels to offset around each pixel with the
            same color

    Returns: None, vector is modified in place
    """

    # For robustness purposes we may want to color more than 1 pixel
    spread = range(-buffer, 2 + buffer)

    for di, dj in product(spread, spread):
        vector[(stride - i - di) * stride + (j + dj)] = color



import argparse
from inspect import signature
from pathlib import Path

from geograypher.predictors.ortho_segmentor import assemble_tiled_predictions


def parse_args():
    """Parse and return arguements

    Returns:
        argparse.Namespace: Arguments
    """

    # For optional CLI arguments we must provide a default value. In most cases, these parameters
    # already have a default defined at the function level.
    atp_params = signature(assemble_tiled_predictions).parameters
    # Build the description string
    description = (
        "This entrypoint is used for aggregating predictions on tiled data into a single geospatial "
        + "raster. "
        + "All the arguments are passed to geograypher.entrypoints.assemble_tiled_predictions "
        + "which has the following documentation:\n\n"
        + assemble_tiled_predictions.__doc__
    )

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter, description=description
    )
    parser.add_argument("raster_file", type=Path)
    parser.add_argument("pred_folder", type=Path)
    parser.add_argument("class_savefile", type=Path)

    parser.add_argument("num_classes", type=int)
    parser.add_argument("--counts-savefile", type=Path)
    parser.add_argument(
        "--downweight-edge-frac",
        type=float,
        default=atp_params["downweight_edge_frac"].default,
    )
    parser.add_argument(
        "--nodataval",
        type=int,
        default=atp_params["nodataval"].default,
    )
    # TODO consider adding the argument for count_dtype
    # the challenge here is we need to parse a class type (ex. np.uint8, np.uint16)
    parser.add_argument(
        "--max-overlapping-tiles",
        type=int,
        default=atp_params["max_overlapping_tiles"].default,
    )

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    # Parse the args
    args = parse_args()
    # Unpack them to the function
    assemble_tiled_predictions(**args.__dict__)



"""
Multiview tree detection and triangulation from aerial imagery.
"""

import argparse
import logging
from pathlib import Path
from typing import Optional

import geopandas as gpd
import matplotlib.cm as cm
import numpy as np
import pandas
import pyproj
import pyvista as pv
from matplotlib.colors import Normalize
from shapely.geometry import Point
from tqdm import tqdm

from geograypher.cameras import MetashapeCameraSet
from geograypher.constants import LAT_LON_CRS
from geograypher.meshes.meshes import TexturedPhotogrammetryMesh
from geograypher.predictors.derived_segmentors import RegionDetectionSegmentor
from geograypher.utils.files import ensure_folder
from geograypher.utils.geometric import get_scale_from_transform
from geograypher.utils.visualization import merge_cylinders

TRANSFORMS = {
    None: None,
    "square": lambda x: x**2,
    "cube": lambda x: x**3,
}


logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


def parse_args():
    parser = argparse.ArgumentParser(
        description="Triangulate tree locations from image segmentations.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--images-dir",
        type=Path,
        required=True,
        help="Directory containing raw images or nested images",
    )
    parser.add_argument(
        "--detections-dir",
        type=Path,
        required=True,
        help="Directory containing .gpkg files (one per image), should"
        " match the nested structure of --images-dir",
    )
    parser.add_argument(
        "--camera-file",
        type=Path,
        required=True,
        help="Path to XML file containing camera calibrations and positions"
        " from photogrammetry software",
    )
    parser.add_argument(
        "--mesh-file",
        type=Path,
        required=True,
        help="Path to georeferenced mesh file",
    )
    parser.add_argument(
        "--mesh-crs",
        type=pyproj.crs.CRS.from_epsg,
        required=True,
        help="The CRS to interpret the mesh in (integer)",
    )
    parser.add_argument(
        "--output-dir",
        type=Path,
        required=True,
        help="Output directory for triangulated tree locations",
    )
    parser.add_argument(
        "--original-image-folder",
        type=Path,
        help="If provided, this will be subtracted off the beginning of absolute image paths"
        " stored in the camera_file. See MetashapeCameraSet for details",
    )
    parser.add_argument(
        "--image-extension", type=str, default=".JPG", help="Image file extension"
    )
    parser.add_argument(
        "--similarity-threshold-meters",
        type=float,
        default=4.0,
        help="Ray intersection threshold in meters",
    )
    parser.add_argument(
        "--louvain-resolution",
        type=float,
        default=2.0,
        help="Louvain resolution parameter, larger value = smaller communities",
    )
    parser.add_argument(
        "--nonlinearity",
        choices=TRANSFORMS.keys(),
        default=None,
        help="Nonlinear transform to apply to distances before inverting and"
        " creating the graph weights. The default graph weight is 1/x, where x"
        " is the intersection distance, but with --nonlinearity square the"
        " graph weight would instead be 1/x**2.",
    )
    parser.add_argument(
        "--vis",
        action="store_true",
        help="Generate 3D visualizations for detected trees and rays in --output-dir",
    )
    args = parser.parse_args()

    # Check inputs
    assert args.images_dir.is_dir(), f"Images dir {args.images_dir} doesn't exist"
    assert args.detections_dir.is_dir(), f"Gpkg dir {args.detections_dir} doesn't exist"
    assert args.camera_file.is_file(), f"Camera XML {args.camera_file} doesn't exist"
    assert args.mesh_file.is_file(), f"Mesh {args.mesh_file} doesn't exist"

    # Ensure output directory exists
    ensure_folder(args.output_dir)

    return args


def vis_lines(logger, line_segments_file, communities_file, out_dir, batch=250):

    data = np.load(line_segments_file)
    ray_starts = data["ray_starts"]
    ray_ends = data["ray_ends"]

    data = np.load(communities_file)
    community_IDs = data["ray_IDs"]
    community_points = data["community_points"]

    norm = Normalize(vmin=np.nanmin(community_IDs), vmax=np.nanmax(community_IDs))
    cmap = cm.get_cmap("tab20")
    cylinder_polydata = None
    n_batches = int(np.ceil(len(community_IDs) / batch))
    for i in tqdm(range(n_batches), desc="Building cylinders"):
        islice = slice(i * batch, min((i + 1) * batch, len(community_IDs)))
        batched = merge_cylinders(
            starts=ray_starts[islice],
            ends=ray_ends[islice],
            community_IDs=community_IDs[islice],
            cmap=cmap,
            norm=norm,
        )
        if cylinder_polydata is None:
            cylinder_polydata = batched
        else:
            cylinder_polydata = cylinder_polydata.merge(batched)

    if cylinder_polydata is not None:
        path = out_dir / "rays.ply"
        logger.info(f"Saving visualized cylinders to {path}")
        cylinder_polydata.save(path, texture="RGB")

    cube_polydata = None
    for comm_id, pt in enumerate(tqdm(community_points, desc="Building points")):
        cube = pv.Cube(center=pt, x_length=0.2, y_length=0.2, z_length=0.2)
        color = (np.array(cmap(norm(comm_id)))[:3] * 255).astype(np.uint8)
        cube.point_data["RGB"] = np.tile(color, (cube.n_points, 1))
        if cube_polydata is None:
            cube_polydata = cube
        else:
            cube_polydata = cube_polydata.merge(cube)

    if cube_polydata is not None:
        path = out_dir / "points.ply"
        logger.info(f"Saving visualized cubes to {path}")
        cube_polydata.save(path, texture="RGB")


def multiview_detections(
    images_dir: Path,
    detections_dir: Path,
    camera_file: Path,
    mesh_file: Path,
    mesh_crs: pyproj.crs.CRS,
    output_dir: Path,
    original_image_folder: Optional[Path],
    image_file_extension: str = ".JPG",
    similarity_threshold_meters: float = 0.1,
    louvain_resolution: float = 2.0,
    transform=None,
    vis: bool = False,
):
    """
    Triangulate tree locations from multi-view image segmentations.

    Args:
        images_dir (Path): Directory containing the raw images or nested images
        detections_dir (Path): Directory containing GeoPackage (.gpkg) files with object detection
            results, one file per corresponding image
        camera_file (Path): Path to XML file containing camera calibrations and positions
            from photogrammetry software
        mesh_file (Path): Path to the georeferenced 3D mesh file
        mesh_crs (pyproj.crs.CRS): Coordinate reference system for interpreting the mesh
        output_dir (Path): Directory where results will be saved
        original_image_folder (Optional[Path]): If provided, this will be subtracted off
            the beginning of absolute image paths stored in the --camera-file.
            See MetashapeCameraSet for details
        image_file_extension (str, optional): File extension for images. Defaults to ".JPG"
        similarity_threshold_meters (float, optional): Distance threshold in meters for
            considering ray intersections as similar. Defaults to 0.1
        louvain_resolution (float, optional): Resolution parameter for Louvain community
            detection clustering. Larger values create smaller clusters. Defaults to 2.0
        transform (callable, optional): Nonlinear transformation function to apply to
            intersection distances before creating graph weights. Defaults to None
        vis (bool, optional): Whether to generate 3D visualization files
            for detected trees and ray intersections. Defaults to False

    Returns:
        None: Results are saved to files in the output_dir:
            - tree_locations.gpkg: GeoDataFrame with triangulated tree positions in lat/lon
            - boundary_ceiling.ply and boundary_floor.ply: 3D mesh boundaries
            - rays.ply and points.ply: 3D visualizations (if vis=True)

    Raises:
        AssertionError: If input directories or files don't exist
    """

    logger = logging.getLogger(__name__)

    # Load camera set
    camera_set = MetashapeCameraSet(
        camera_file=camera_file,
        image_folder=images_dir,
        original_image_folder=original_image_folder,
        validate_images=True,
    )
    local_to_epsg_4978 = camera_set.get_local_to_epsg_4978_transform()

    # Load mesh in the photogrammetry reference frame (PRF). This is because the
    # camera locations are defined in the PRF
    mesh = TexturedPhotogrammetryMesh(mesh_file, input_CRS=mesh_crs)
    # Convert the mesh into one in the same coordinate frame as the camera set
    mesh.get_mesh_in_cameras_coords(camera_set, inplace=True)

    # Create boundary layers between the ground and the treetops that we
    # will check for ray intersections between. Note that the z_buffer
    # defines, in meters (which is then translated to local scale), the
    # spacing of the covering meshes from the mesh model. For example,
    # z_buffer[0] is the spacing between mesh and floor, and z_buffer[1]
    # is the spacing between mesh and ceiling.
    local_scale = 1 / get_scale_from_transform(local_to_epsg_4978)
    ceiling, floor = mesh.export_covering_meshes(
        N=50, z_buffer=(0, 1 * local_scale), subsample=2
    )
    mesh.save_mesh(output_dir / "mesh-local.ply")
    ceiling.save(output_dir / "boundary_ceiling.ply")
    floor.save(output_dir / "boundary_floor.ply")
    logger.info("Boundary meshes saved")

    # Load region detector
    logger.info("Loading region detection segmentor")
    detector = RegionDetectionSegmentor(
        base_folder=images_dir,
        lookup_folder=detections_dir,
        label_key=None,
        class_map=None,
    )

    # Triangulate detections (tree locations) and add lines/points to plotter
    tree_points = camera_set.triangulate_detections(
        detector=detector,
        boundaries=(ceiling, floor),
        limit_ray_length_meters=160,
        limit_angle_from_vert=np.deg2rad(50),
        similarity_threshold_meters=similarity_threshold_meters,
        transform=transform,
        louvain_resolution=louvain_resolution,
        out_dir=output_dir,
    )

    # Visualize the intersections
    if vis:
        vis_lines(
            logger,
            line_segments_file=output_dir / "line_segments.npz",
            communities_file=output_dir / "communities.npz",
            out_dir=output_dir,
        )

    # Save results
    gpkg_path = output_dir / "tree_locations.gpkg"
    # Note that we want to store (lat, lon) - which is what tree_points
    # comes in as - in the form (y, x)
    gdf = gpd.GeoDataFrame(
        geometry=[Point(x, y, z) for y, x, z in tree_points],
        crs=LAT_LON_CRS,
    )
    gdf.to_file(gpkg_path)
    logger.info(f"Saved triangulated tree locations to {gpkg_path}")


if __name__ == "__main__":
    args = parse_args()
    multiview_detections(
        images_dir=args.images_dir,
        detections_dir=args.detections_dir,
        camera_file=args.camera_file,
        mesh_file=args.mesh_file,
        mesh_crs=args.mesh_crs,
        output_dir=args.output_dir,
        original_image_folder=args.original_image_folder,
        image_file_extension=args.image_extension,
        similarity_threshold_meters=args.similarity_threshold_meters,
        louvain_resolution=args.louvain_resolution,
        transform=TRANSFORMS[args.nonlinearity],
        vis=args.vis,
    )



import json
import typing
from itertools import product
from pathlib import Path

import networkx
import numpy as np
from tqdm import tqdm

from geograypher.constants import EARTH_CENTERED_EARTH_FIXED_CRS, LAT_LON_CRS, PATH_TYPE
from geograypher.utils.geospatial import convert_CRS_3D_points


def create_ramped_weighting(
    rectangle_shape: typing.Tuple[int, int], ramp_dist_frac: float
) -> np.ndarray:
    """Create a ramped weighting that is higher toward the center with a max value of 1 at a fraction from the edge

    Args:
        rectangle_shape (typing.Tuple[int, int]): Size of rectangle to create a mask for
        ramp_dist_frac (float): Portions at least this far from an edge will have full weight

    Returns:
        np.ndarray: An array representing the weighting from 0-1
    """
    i_ramp = np.clip(np.linspace(0, 1 / ramp_dist_frac, num=rectangle_shape[0]), 0, 1)
    j_ramp = np.clip(np.linspace(0, 1 / ramp_dist_frac, num=rectangle_shape[1]), 0, 1)

    i_ramp = np.minimum(i_ramp, np.flip(i_ramp))
    j_ramp = np.minimum(j_ramp, np.flip(j_ramp))

    i_ramp = np.expand_dims(i_ramp, 1)
    j_ramp = np.expand_dims(j_ramp, 0)

    ramped_weighting = np.minimum(i_ramp, j_ramp)
    return ramped_weighting


def compute_approximate_ray_intersections(
    a0: np.ndarray, a1: np.ndarray, b0: np.ndarray, b1: np.ndarray, clamp: bool = False
) -> typing.Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Compute closest points and distances between N line segments a0->a1 and
    b0->b1. Returns (N, N, 3), (N, N, 3), (N, N). If clamp is True, then respect
    the line segment ends. If clamp is False, then use the infinite rays.

    Based on https://stackoverflow.com/questions/2824478/shortest-distance-between-two-line-segments

    Args:
        a0 (np.ndarray): Start points of the first segments (N, 3).
        a1 (np.ndarray): End points of the first segments (N, 3).
        b0 (np.ndarray): Start points of the second segments (N, 3).
        b1 (np.ndarray): End points of the second segments (N, 3).
        clamp (bool, optional): If True, the closest points are clamped to the
            segment endpoints. If False, the closest points may be anywhere
            along the infinite rays.

    Returns:
        pA (np.ndarray): Closest point on the A segments (axis 0) compared to
            each of the B segments (axis 1) (N, N, 3). For example, the closest
            point between A[5] and B[2] at pA[5, 2]
        pB (np.ndarray): Closest point on the B segments (axis 1) compared to
            each of the A segments (axis 1) (N, N, 3)
        dist (np.ndarray): The minimum distance between the A (axis 0) and B
            (axis 1) segments.
    """

    # (N, 3) vectors, representing the A and B line segments
    A = a1 - a0
    B = b1 - b0
    # (N,) distances, representing the length of each A and B vector
    magA = np.linalg.norm(A, axis=1)
    magB = np.linalg.norm(B, axis=1)
    # (N, 3) vectors, representing the A and B unit vectors
    _A = A / magA[:, None]
    _B = B / magB[:, None]

    # Expand the A vectors to (N, 1, 3) and the B vectors to (1, N, 3) so that
    # they project together to an (N, N, 3) matrix later
    a0_exp = a0[:, None, :]
    b0_exp = b0[None, :, :]
    _A_exp = _A[:, None, :]
    _B_exp = _B[None, :, :]

    # (N, N, 3) cross product. The cross product of unit vectors calculates the
    # area of the parallelogram formed by the unit vectors, a larger value
    # indicates lines that are more orthogonal
    cross = np.cross(_A_exp, _B_exp)
    # (N, N) representing the squared area of the unit vector parallelograms
    denom = np.linalg.norm(cross, axis=2) ** 2

    # (N, N, 3) matrix, representing the vector from the start of each A
    # segment to the start of each B segment
    t = b0_exp - a0_exp

    # (N, N) matrix, calculate the determinant by multiplying all axes (ijk)
    # and then sum across k, leaving only an ij matrix. The determinant
    # represents how a given transform scales space
    detA = np.einsum("ijk,ijk->ij", np.cross(t, _B_exp), cross)
    detB = np.einsum("ijk,ijk->ij", np.cross(t, _A_exp), cross)

    # Make the denom safe for division by replacing 0 values with 1. Later we
    # check for parallel locations and fix them
    parallel = denom == 0
    denom[parallel] = 1

    # t0 and t1 are (N, N) matrices representing how far along A the closest
    # point is (t0) where 0 is at a0 and 1 is at a1. B and t1 are the same.
    t0 = detA / denom
    t1 = detB / denom

    if clamp:
        # (N, N) matrix where t0 and t1 are clipped to the vector length
        t0_clamped = np.clip(t0, 0, magA[:, None])
        t1_clamped = np.clip(t1, 0, magB[None, :])

        # (N, N, 3) matrix where we start at each vector starting point (a0
        # and b0 expanded) then travel along them to the closest clamped point
        pA = a0_exp + t0_clamped[:, :, None] * _A_exp
        pB = b0_exp + t1_clamped[:, :, None] * _B_exp

        # Check whether the original scale values where out of bounds
        oob_A = (t0 < 0) | (t0 > magA[:, None])
        oob_B = (t1 < 0) | (t1 > magB[None, :])

        # Broadcast a0 and _A to (N, N, 3), similar for b0/_B
        a0_bcast = np.broadcast_to(a0[:, None, :], pA.shape)
        _A_bcast = np.broadcast_to(_A[:, None, :], pA.shape)
        b0_bcast = np.broadcast_to(b0[None, :, :], pB.shape)
        _B_bcast = np.broadcast_to(_B[None, :, :], pB.shape)

        # If any of the scale vectors were clipped, we may have to adjust the
        # closest point on the opposite vector. For example if the closest
        # point on an A vector was clipped, the corresponding B point may need
        # to update
        if np.any(oob_A):
            # Get the projection of the new A point onto the B vectors
            dot = np.einsum("ijk,ijk->ij", pA - b0_bcast, _B_bcast)
            # Get a new B point matching that projection
            dot_clipped = np.clip(dot, 0, magB[None, :])
            pB[oob_A] = b0_bcast[oob_A] + dot_clipped[oob_A, None] * _B_bcast[oob_A]
        if np.any(oob_B):
            # Get the projection of the new B point onto the A vectors
            dot = np.einsum("ijk,ijk->ij", pB - a0_bcast, _A_bcast)
            # Get a new A point matching that projection
            dot_clipped = np.clip(dot, 0, magA[:, None])
            pA[oob_B] = a0_bcast[oob_B] + dot_clipped[oob_B, None] * _A_bcast[oob_B]
    else:
        # If there is no clamping, we use the scale value as-is to get the
        # closest points between the rays
        pA = a0_exp + t0[:, :, None] * _A_exp
        pB = b0_exp + t1[:, :, None] * _B_exp

    # Handle parallel case
    if np.any(parallel):
        # Get the (N, N) projection of each point in b0 onto each unit vector
        # in A. Subtract the (N, 1) projection of each point in a0 along their
        # unit vector. By subtracting them, we calculating whether b0 falls
        # "ahead" or "behind" of a0, along the relevant axis (_A)
        d0 = np.einsum("ij,kj->ik", _A, b0) - np.einsum("ij,ij->i", _A, a0).reshape(
            -1, 1
        )
        if clamp:
            # Same logic as d0, but with a1 and b1
            d1 = np.einsum("ij,kj->ik", _A, b1) - np.einsum("ij,ij->i", _A, a0).reshape(
                -1, 1
            )

            # Check in which of the (N, N) combinations b0 and b1 fall before
            # a0, forming an (N, N) boolean mask
            before = (d0 <= 0) & (d1 <= 0) & parallel
            # Same logic for after
            after = (d0 >= magA[:, None]) & (d1 >= magA[:, None]) & parallel
            # If an A and B vector is parallel and partially overlapping,
            # middle will be true (N, N) boolean mask
            middle = parallel & ~(before | after)

            if np.any(before):
                # Set pA to a0 where relevant
                pA[before] = a0[:, None, :][before]
                # Set pB to whichever endpoint is closer to a0
                pB[before] = np.where(
                    np.abs(d0[before])[:, None] < np.abs(d1[before])[:, None],
                    b0[None, :, :][before],
                    b1[None, :, :][before],
                )

            if np.any(after):
                # Set pA to a1 where relevant
                pA[after] = a1[:, None, :][after]
                # Set pB to whichever endpoint is closer to a1
                pB[after] = np.where(
                    np.abs(d0[after])[:, None] < np.abs(d1[after])[:, None],
                    b0[None, :, :][after],
                    b1[None, :, :][after],
                )

            if np.any(middle):
                # Clip d0 along the A vectors (from 0 to the magnitude of A)
                t_mid = np.clip(
                    d0[middle], 0, np.broadcast_to(magA[:, None], d0.shape)[middle]
                )

                # Get the broadcast components that align with the state we
                # care about (middle)
                a0_bcast = np.broadcast_to(a0[:, None, :], pA.shape)
                _A_bcast = np.broadcast_to(_A[:, None, :], pA.shape)
                a0_mid = a0_bcast[middle]
                _A_mid = _A_bcast[middle]

                # Reproject onto the A vectors
                pA[middle] = a0_mid + t_mid[:, None] * _A_mid

                # Get the vector from the new pA points to b0
                b0_bcast = np.broadcast_to(b0[None, :, :], pB.shape)
                a2b = b0_bcast[middle] - pA[middle]

                # Then subtract the component that is along the A vectors, and
                # that gives us the final pB location
                alongA = np.einsum("ij,ij->i", a2b, _A_mid)[:, None] * _A_mid
                pB[middle] = pA[middle] + (a2b - alongA)

        else:
            # If we're not clamping, arbitrarily set the parallel "closest
            # points" to b0 and the matching point on A
            a0_bcast = np.broadcast_to(a0[:, None, :], pA.shape)
            b0_bcast = np.broadcast_to(b0[None, :, :], pB.shape)
            _A_bcast = np.broadcast_to(_A[:, None, :], pA.shape)
            pA[parallel] = (
                a0_bcast[parallel] + d0[:, :, None][parallel] * _A_bcast[parallel]
            )
            pB[parallel] = b0_bcast[parallel]

    # pA and pB are existing (N, N, 3) matrices. In addition, calculate the
    # (N, N) distances between each pair
    return pA, pB, np.linalg.norm(pA - pB, axis=2)


def triangulate_rays_lstsq(starts, directions):
    # https://github.com/tensorflow/graphics/blob/master/tensorflow_graphics/geometry/representation/ray.py#L284-L371

    # Ensure that we have unit vectors specifying the direction of the ray

    # Build a cross product matrix for each direction, this is the left side
    # The right side is the dot product between this and the start ray

    As = []
    bs = []
    # TODO build this all at once rather than iteratively
    for start, direction in zip(starts, directions):
        x, y, z = direction
        # https://math.stackexchange.com/questions/3764426/matrix-vector-multiplication-cross-product-problem
        cross_matrix = np.array(
            [
                [0, -z, y],
                [z, 0, -x],
                [-y, x, 0],
            ]
        )
        As.append(cross_matrix)
        b = cross_matrix @ start
        bs.append(b)

    A = np.concatenate(As, axis=0)
    b = np.concatenate(bs, axis=0)

    x, _, _, _ = np.linalg.lstsq(A, b, rcond=None)
    return x


def compute_3D_triangle_area_vectorized(corners: np.ndarray, return_z_proj_area=True):
    """_summary_

    Args:
        corners (np.ndarray): (n_faces, n)
        return_z_proj_area (bool, optional): _description_. Defaults to True.

    Returns:
        _type_: _description_
    """
    A, B, C = corners
    # https://math.stackexchange.com/questions/2152754/calculate-3d-triangle-area-by-determinant
    u = B - A
    v = C - A

    # Save for future computation
    u0v1_min_u1v0 = u[0] * v[1] - u[1] * v[0]
    area = (
        1
        / 2
        * np.sqrt(
            np.power(u[1] * v[2] - u[2] * v[1], 2)
            + np.power(u[2] * v[0] - u[0] * v[2], 2)
            + np.power(u0v1_min_u1v0, 2)
        )
    )

    if return_z_proj_area:
        area_z_proj = np.abs(u0v1_min_u1v0) / 2
        return area, area_z_proj

    return area


def compute_3D_triangle_area(corners, return_z_proj_area=True):
    A, B, C = corners
    # https://math.stackexchange.com/questions/2152754/calculate-3d-triangle-area-by-determinant
    u = B - A
    v = C - A

    # Save for future computation
    u0v1_min_u1v0 = u[0] * v[1] - u[1] * v[0]
    area = (
        1
        / 2
        * np.sqrt(
            np.power(u[1] * v[2] - u[2] * v[1], 2)
            + np.power(u[2] * v[0] - u[0] * v[2], 2)
            + np.power(u0v1_min_u1v0, 2)
        )
    )

    if return_z_proj_area:
        area_z_proj = np.abs(u0v1_min_u1v0) / 2
        return area, area_z_proj

    return area


def intersection_average(starts: np.ndarray, ends: np.ndarray) -> np.ndarray:
    """
    Given arrays of line segment start and end points, compute the average of the closest
    intersection points between all pairs of segments.

    Args:
        starts (np.ndarray): (N, 3) array of segment start points
        ends (np.ndarray): (N, 3) array of segment end points

    Returns:
        np.ndarray: (3,) array, the average intersection point
    """
    closest_points = []
    pA, pB, _ = compute_approximate_ray_intersections(
        a0=starts, a1=ends, b0=starts, b1=ends, clamp=True
    )
    mask = ~np.eye(starts.shape[0], dtype=bool)
    return np.mean(np.vstack([pA[mask], pB[mask]]), axis=0)


def chunk_slices(
    N: int, step: int
) -> typing.Iterator[typing.Tuple[slice, slice, bool]]:
    """
    Yield slices for (step, step) chunks of the upper triangular area
    of an (N, N) square matrix.

    For example, if N=5 and step=2, the slices would grab:
        1 1 2 2 3
        1 1 2 2 3
        - - 4 4 5
        - - 4 4 5
        - - - - 6
    And in the (1, 4, 6) cases, is_diag would be True

    Yields:
        Each yielded value is (islice, jslice, is_diag), where:
        - islice: slice along the first axis
        - jslice: slice along the second axis
        - is_diag: True if the chunk is on the diagonal (i == j)
    """
    ranges = range(0, N, step)
    for i, j in product(ranges, repeat=2):
        if j >= i:  # upper triangle including diagonal
            islice = slice(i, min(i + step, N))
            jslice = slice(j, min(j + step, N))
            yield islice, jslice, i == j


def format_graph_edges(
    islice: slice,
    jslice: slice,
    dist: np.ndarray,
    ray_IDs: np.ndarray,
) -> typing.List[typing.Tuple[int, int, typing.Dict[str, float]]]:
    """
    This function generates edge definitions for a graph where nodes represent rays and
    edges represent valid intersections between rays. It applies three filtering criteria:
    1. Only uses edges where (i, j) is finite (not NaN)
    2. Only uses edges where i < j (keeps it upper triangular)
    3. Only uses edges between rays from different images (different ray_IDs)

    Args:
        islice (slice): Slice indicating the block of rows being processed in the
            chunked computation
        jslice (slice): Slice indicating the block of columns being processed in
            the chunked computation
        dist (np.ndarray): Distance matrix containing distances between rays (chunked)
            This is the [islice, jslice] section of a larger distance matrix.
        ray_IDs (np.ndarray): Array of identifiers indicating which image each ray comes
            from (not chunked)

    Returns:
        List[Tuple[int, int, Dict[str, float]]]: List of edge definitions, where each edge
        is a tuple of
        - i index in the adjacency matrix
        - j index in the adjacency matrix
        - weight_dict of the form {"weight": value}
    """

    # The places where the array is finite are the valid graph distances
    i_inds, j_inds = np.where(np.isfinite(dist))

    # Pre-calculate the inverse distance
    weights = 1 / dist

    return [
        (
            int(i) + islice.start,
            int(j) + jslice.start,
            {"weight": float(weights[i, j])},
        )
        for i, j in zip(i_inds, j_inds)
        if (i + islice.start < j + jslice.start)
        and (ray_IDs[i + islice.start] != ray_IDs[j + jslice.start])
    ]


def calc_graph_weights(
    starts: np.ndarray,
    ends: np.ndarray,
    ray_IDs: np.ndarray,
    similarity_threshold: float,
    out_dir: typing.Optional[PATH_TYPE] = None,
    min_dist: float = 1e-6,
    step: int = 5000,
    transform: typing.Optional[typing.Callable[[np.ndarray], np.ndarray]] = None,
) -> typing.Union[Path, typing.List[typing.Tuple]]:
    """
    This function processes sets of ray segments to build a graph where edges represent
    ray intersections. The weight of each edge is inversely proportional to the
    intersection distance between rays. For memory efficiency with large numbers of
    segments, the computation is done in chunks.

    Args:
        starts (np.ndarray): (N, 3) array of ray start points
        ends (np.ndarray): (N, 3) array of ray end points
        ray_IDs (np.ndarray): (N,) array of integers identifying which image
            each ray comes from
        similarity_threshold (float): Maximum intersection distance to consider when
            creating graph edges. Greater distances will be dropped.
        out_dir (PATH_TYPE, optional): Directory to save the output JSON file containing
            edge information. If None, no file is saved and a list of edge weights
            is returned instead.
        min_dist (float, optional): Minimum intersection distance to allow, used to
            avoid division by zero when calculating weights. Defaults to 1e-6
        step (int, optional): Number of rays to process in each chunk to manage memory
            usage. Defaults to 5000
        transform (callable, optional): Function to apply to distances before inversion
            for weight calculation. For example, if you want graph weights to be
            distance^3, use lambda x: x**3. Defaults to None

    Returns:
        Union[Path, List[Tuple]]: If out_dir is provided, returns the path to the JSON
        file containing the edge information. Otherwise returns a list of tuples, each
        containing (start_idx, end_idx, weight_dict) where weight_dict contains the
        edge weight information
    """

    # Calculate and filter intersection distances
    # For memory reasons, we need to iterate over blocks. When the number of segments starts
    # getting very large, the matrices for calculating ray intersections take a great
    # deal of RAM
    edge_weights = []
    num_steps = np.ceil(len(starts) / step)
    total = int(num_steps * (num_steps + 1) / 2)
    for islice, jslice, diagonal in tqdm(
        chunk_slices(N=len(starts), step=step),
        total=total,
        desc="Calculating graph weights",
    ):
        _, _, dist = compute_approximate_ray_intersections(
            a0=starts[islice],
            a1=ends[islice],
            b0=starts[jslice],
            b1=ends[jslice],
            clamp=True,
        )
        if diagonal:
            np.fill_diagonal(dist, np.nan)
        dist[dist > similarity_threshold] = np.nan
        dist[dist < min_dist] = min_dist

        # Apply transform if provided
        if transform is not None:
            dist = transform(dist)

        # Create edge weights for valid intersections
        edge_weights.extend(format_graph_edges(islice, jslice, dist, ray_IDs))

    if out_dir is None:
        return edge_weights
    else:
        path = Path(out_dir) / "edge_weights.json"
        with path.open("w") as file:
            json.dump(edge_weights, file)
        return path


def calc_communities(
    starts: np.ndarray,
    ends: np.ndarray,
    edge_weights: typing.List[typing.Tuple[int, int, typing.Dict[str, float]]],
    louvain_resolution: float = 1.0,
    out_dir: typing.Optional[PATH_TYPE] = None,
    transform_to_epsg_4978: typing.Optional[np.ndarray] = None,
) -> typing.Union[Path, typing.Dict[str, np.ndarray]]:
    """
    Build a networkx graph from adjacency information. Each node represents
    a detection while the edges represent the quality of the matches between detections.

    Args:
        starts (np.ndarray): (N, 3) array of ray start points
        ends (np.ndarray): (N, 3) array of ray end points
        edge_weights (List[Tuple[int, int, Dict[str, float]]]): List of edges defining
            the graph connectivity. Each edge is (start_idx, end_idx, weight_dict) where
            weight_dict contains the edge weight information
        louvain_resolution (float): Resolution hyperparameter for the Louvain community
            detection algorithm. Higher values lead to more communities
        out_dir (PATH_TYPE, optional): Directory to save the output NPZ file containing
            community information. If None, results are returned as a dictionary
        transform_to_epsg_4978 (np.ndarray, optional): 4x4 transformation matrix to
            convert points from local coordinates to EPSG:4978 (Earth-centered Earth-fixed)

    Returns:
        Union[Path, Dict[str, np.ndarray]]: If out_dir is provided, returns the path to
            the NPZ file containing the community information. Otherwise returns a
            dictionary with keys:
            - 'ray_IDs': (N,) array mapping each ray to its community ID
            - 'community_points': (M, 3) array of 3D points representing each community
            - 'community_points_latlon': (M, 3) array of lat/lon points (only if
               transform_to_epsg_4978 is provided)
    """

    # Build up the basic graph from edge weights
    graph = networkx.Graph(edge_weights)

    # Check that the graph is non empty
    if len(graph) > 0:

        # Determine Louvain communities which are sets of nodes. Ideally, each
        # community represents a set of detections that correspond to one 3D object
        communities = networkx.community.louvain_communities(
            graph, weight="weight", resolution=louvain_resolution
        )
        # Sort the communities by size
        communities = sorted(communities, key=len, reverse=True)

        # Triangulate the rays for each community to identify the 3D location
        community_points = []
        # Record the community IDs per ray
        num_rays = starts.shape[0]
        ray_IDs = np.full(num_rays, fill_value=np.nan)
        # Iterate over communities
        for community_ID, community in enumerate(
            tqdm(communities, desc="Build community points")
        ):
            # Get the ray indices that belong to that community
            community_indices = np.array(list(community))
            # Record the community ID for the corresponding rays
            ray_IDs[community_indices] = community_ID
            # Use the average of the closest points between rays as the
            # representative point for the community
            community_points.append(
                intersection_average(
                    starts=starts[community_indices],
                    ends=ends[community_indices],
                )
            )

        # Stack all of the points into one vector
        community_points = np.vstack(community_points)

        result = {
            "ray_IDs": ray_IDs,
            "community_points": community_points,
        }

        if transform_to_epsg_4978 is not None:
            # Append a column of all ones to make the homogenous coordinates
            homogenous = np.concatenate(
                [community_points, np.ones_like(community_points[:, 0:1])],
                axis=1,
            )
            # Use the transform matrix to transform the points into the earth
            # centered, earth fixed frame (EPSG:4978)
            community_points_epsg_4978 = (transform_to_epsg_4978 @ homogenous.T).T
            # Convert the points from earth centered, earth fixed frame to lat lon
            community_points_lat_lon = convert_CRS_3D_points(
                community_points_epsg_4978,
                input_CRS=EARTH_CENTERED_EARTH_FIXED_CRS,
                output_CRS=LAT_LON_CRS,
            )
            result["community_points_latlon"] = community_points_lat_lon

    else:
        # Handle the empty case
        result = {
            "ray_IDs": np.zeros((0,), dtype=int),
            "community_points": np.zeros((0, 3)),
        }
        if transform_to_epsg_4978 is not None:
            result["community_points_latlon"] = np.zeros((0, 3))

    if out_dir is not None:
        path = Path(out_dir) / "communities.npz"
        np.savez(path, **result)
        return path
    else:
        return result


def fair_mode_non_nan(values: np.ndarray) -> np.ndarray:
    """
    Compute the most common value per row in an array of integers and nans. This behaves similarly to
    scipy.stats.mode(values, axis=1, nan_policy="omit") except that for values with equal counts,
    one is chosen randomly, rather than taking the lower value.

    Args:
        values (np.ndarray): (n, m) The input values (float-typed), consisting of integers and nans

    Returns:
        np.ndarray: (n,) the most common value per row
    """
    max_val = np.nanmax(values)
    # All input values are nan, return all nans
    if np.isnan(max_val):
        return np.full((values.shape[0],), fill_value=np.nan)

    max_val = int(max_val)
    # TODO consider using unique if these indices are sparse
    counts_per_value_per_row = np.array(
        [np.sum(values == i, axis=1) for i in range(max_val + 1)]
    ).T
    # Check which entires had no classes reported and mask them out
    # TODO consider removing these rows beforehand
    zeros_mask = np.all(counts_per_value_per_row == 0, axis=1)
    # We want to fairly tiebreak since np.argmax will always take th first index
    # This is hard to do in a vectorized way, so we just add a small random value
    # independently to each element
    counts_per_value_per_row = (
        counts_per_value_per_row
        + np.random.random(counts_per_value_per_row.shape) * 0.5
    )
    most_common_value_per_row = np.argmax(counts_per_value_per_row, axis=1).astype(
        float
    )
    # Set any faces with zero counts to nan
    most_common_value_per_row[zeros_mask] = np.nan
    return most_common_value_per_row



from pathlib import Path
from typing import Dict, List

from geograypher.cameras import MetashapeCameraSet, PhotogrammetryCameraSet
from geograypher.constants import PATH_TYPE
from geograypher.utils.image import rotate_by_roll_pitch_yaw


def create_rig_cameras_from_equirectangular(
    camera_file: PATH_TYPE,
    original_images: PATH_TYPE,
    perspective_images: PATH_TYPE,
    rig_camera: Dict[str, float],
    rig_orientations: List[Dict[str, float]],
    perspective_filename_format_str: str,
) -> PhotogrammetryCameraSet:
    """
    Create a synthetic camera set corresponding to a rig of perspective cameras derived from an
    equirectangular camera. This is a workaround for directly projecting an equirectangular
    representation because 1) the rendering libraries do not support equirectangular models and 2)
    it's often easier to do prediction tasks on a perspective crop.

    Args:
        camera_file (PATH_TYPE):
            Path to a metashape results file from the real equirectangular images.
        original_images (PATH_TYPE):
            The location of equirectangular images used to generate the photogrammetry products
        perspective_images (PATH_TYPE):
            The resampled perspective images obtained from the equirectangular images. It's expected
            that each image filename contains information about the original image it was generated
            from and then the orientation it was captured in.
        rig_camera (Dict[str, float]):
            A dict with "f", "cx", "cy", "image_width" and "image_height" keys. This defines the
            parameters of the synthetic camera. Note that since this is synthetic, it's common for
            cx and cy to be zero.
        rig_orientations (List[Dict[str, float]]):
            A list of dicts where each one has the keys "roll_deg", "pitch_deg", and "yaw_deg".
            Each dict indicates the rotation of a given camera in the rig. The center of the
            equirectangular image corresponds to zero roll.
        perspective_filename_format_str (str):
            A format string that accepts named arguments "roll_deg", "pitch_deg", and "yaw_deg". The
            result of this string should be concatenated to the original equirectangular image
            filename (before the extension) to obtain the filename of the perspective image.

    Returns:
        PhotogrammetryCameraSet: A camera set representing the synthetic rig.
    """
    # Load the camera file. This just serves as an easy way to parse the required information,
    # it will not be used directly. The sensor params are defaulted since the spherical camera does
    # not have these attributes but they are required for loading the data.
    initial_camera_set = MetashapeCameraSet(
        camera_file=camera_file,
        image_folder=perspective_images,
        original_image_folder=original_images,
        default_sensor_params={"f": 1.0, "cx": 0.0, "cy": 0.0},
    )

    # Extract the attributes from the original camera set
    cam_to_world_transforms = [
        c.cam_to_world_transform for c in initial_camera_set.cameras
    ]
    image_filenames = [c.image_filename for c in initial_camera_set.cameras]

    # Compute the rotation transforms for each of the cameras in the rig
    rig_transforms = [
        rotate_by_roll_pitch_yaw(**rig_orientation, return_4x4=True)
        for rig_orientation in rig_orientations
    ]

    # Compute the extension string for each image
    image_extensions = [
        perspective_filename_format_str.format(**rig_orientation)
        for rig_orientation in rig_orientations
    ]

    # Apply each of the rig transforms to each transform in the cameras set
    new_transforms = [
        cam_to_world @ rig_transform
        for cam_to_world in cam_to_world_transforms
        for rig_transform in rig_transforms
    ]
    # Determine the new image paths
    new_image_filenames = [
        Path(
            image_filename.parent,
            image_filename.stem + image_extension + ".png",
        )
        for image_filename in image_filenames
        for image_extension in image_extensions
    ]
    # TODO consider supporting multiple sensors within a rig. Though it's a synthetic rig so that
    # might not ever be needed.
    sensor_ids = [0] * len(new_image_filenames)

    # Use this information to create a new camera set that has one element for each camera in the
    # synthetic rig.
    rig_cameras = PhotogrammetryCameraSet(
        cam_to_world_transforms=new_transforms,
        intrinsic_params_per_sensor_type={0: rig_camera},
        image_filenames=new_image_filenames,
        sensor_IDs=sensor_ids,
        local_to_epsg_4978_transform=initial_camera_set.get_local_to_epsg_4978_transform(),
    )

    return rig_cameras



from typing import Optional, Tuple

import numpy as np
import piexif
from PIL import Image
from scipy.spatial.transform import Rotation
from skimage.transform import downscale_local_mean, warp


def get_GPS_exif(filename):
    im = Image.open(filename)
    exif_dict = piexif.load(im.info["exif"])
    lat = exif_dict["GPS"][piexif.GPSIFD.GPSLatitude]
    lon = exif_dict["GPS"][piexif.GPSIFD.GPSLongitude]
    lat = (
        lat[0][0] / lat[0][1]
        + lat[1][0] / (lat[1][1] * 60)
        + lat[2][0] / (lat[2][1] * 3600)
    )
    lon = (
        lon[0][0] / lon[0][1]
        + lon[1][0] / (lon[1][1] * 60)
        + lon[2][0] / (lon[2][1] * 3600)
    )
    # https://stackoverflow.com/questions/54196347/overwrite-gps-coordinates-in-image-exif-using-python-3-6
    return lon, lat


def rotate_by_roll_pitch_yaw(
    roll_deg: float,
    pitch_deg: float,
    yaw_deg: float,
    return_4x4: bool = False,
) -> np.ndarray:
    # Convert from degrees to radians
    yaw = np.deg2rad(yaw_deg)
    pitch = np.deg2rad(pitch_deg)
    roll = np.deg2rad(roll_deg)

    # RPY is defined with roll about X, pitch about Y, yaw about Z. We need to convert from the
    # perspective camera frame axes convention to the RPY convention.
    # New Z has to be old -Y
    # New Y has to be old X
    # New X has to be old Z
    permutation_matrix = np.array([[0, 0, 1], [1, 0, 0], [0, -1, 0]])

    # Calculate the rotation matrix corresponding to the roll-pitch-yaw convention
    # https://stackoverflow.com/questions/74434119/scipy-rotation-matrix-from-as-euler-angles
    rotation_matrix = Rotation.from_euler("ZYX", [yaw, pitch, roll]).as_matrix()

    # The permutation matrix can be thought of first converting into the conventional RPY frame,
    # then applying the rotation, then converting back into the camera frame.
    rotation_matrix_in_cam_frame = (
        permutation_matrix.T @ rotation_matrix @ permutation_matrix
    )

    if return_4x4:
        # Pad with zeros except for a 1 at 3,3
        rotation_matrix_in_cam_frame = np.concatenate(
            [
                np.concatenate(
                    [rotation_matrix_in_cam_frame, np.zeros((3, 1))], axis=1
                ),
                np.array([[0, 0, 0, 1]]),
            ],
            axis=0,
        )

    return rotation_matrix_in_cam_frame


def flexible_inputs_warp(
    input_image: np.ndarray,
    inverse_map: np.ndarray,
    interpolation_order: int = None,
    fill_value: float = 0.0,
) -> np.ndarray:
    """
    Extends the functionality of skimage.transform.warp to handle arbitrary datatypes and
    multi-channel images
    """
    # Temporarily expand grayscale images to be (N, M, 1)
    input_image = np.atleast_3d(input_image)

    # Note if the fill_value is not actually used and is in fact outside the range of the input
    # values it will result in an artificially small range of data for the real values. But it won't
    # affect the results aside from the possible loss of precision.
    # Compute the min and max of values, including both the inputs and the fill value
    input_min = min(np.min(input_image), fill_value)
    input_max = max(np.max(input_image), fill_value)
    # Compute the range of values
    min_max_range = input_max - input_min

    # If there's no variation, we can avoid division by 0 issues and save computation by just
    # returning an array of one value
    if min_max_range == 0:
        return np.full_like(np.squeeze(input_image), fill_value=fill_value)

    # Record the original datatype
    initial_dtype = input_image.dtype
    # Rescale to 0-1
    input_image = (input_image.astype(float) - input_min) / min_max_range
    # The fill value will be applied to the rescaled image, so it must be rescaled similarly
    rescaled_fill_value = (float(fill_value) - input_min) / min_max_range

    # Create an output image that's the shape of inverse map (minus first dimension which is 2 [i,j])
    # and the number of channels from the input image
    output_image = np.zeros(inverse_map.shape[1:] + (input_image.shape[2],))
    # For each color channel, do the warping
    for channel in range(input_image.shape[2]):
        warped = warp(
            image=input_image[:, :, channel],
            inverse_map=inverse_map,
            order=interpolation_order,  # interpolation strategy for fractional pixels
            mode="constant",  # fill unseen areas with cval
            cval=rescaled_fill_value,
            clip=True,  # clip to [0,1]
            preserve_range=True,  # keep original range without rescaling
        )
        output_image[:, :, channel] = warped

    # Convert to the original range and datatype
    output_image = ((output_image * min_max_range) + input_min).astype(initial_dtype)

    # If the original image was grayscale, remove the added singleton dimension
    return np.squeeze(output_image)


def perspective_from_equirectangular(
    equi_img: np.ndarray,
    fov_deg: float,
    output_size: Tuple[int, int] = (1440, 1440),
    yaw_deg: float = 0,
    pitch_deg: float = 0,
    roll_deg: float = 0,
    warp_order: int = 1,
    oversample_factor: int = 1,
    return_mask: bool = False,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Sample a perspective image from an equirectangular (360) image. Different parameters of the
    "virtual camera" can be controlled, including the field of view, orientation (yaw, pitch, roll),
    and output image size.

    The rotations are defined using the roll-pitch-yaw convention. The roll axis corresponds to the
    camera Z (forward) axis, pitch corresponds to the X axis, and yaw corresponds to the Y axis.

    With RPY all set to zero, the camera is looking toward the center of the equirectangular image.

    Args:
        equi_img (np.ndarray):
            Equirectangular image in (H, W, C) format.
        fov_deg float: camera horizontal field of view.
        output_size Tuple[int, int]: Shape of the output image to sample (i, j).
        yaw_deg (float, optional): yaw camera orientation. Defaults to 0.
        pitch_deg (float, optional): pitch camera orientation. Defaults to 0.
        roll_deg (float, optional): roll camera orientation. Defaults to 0.
        warp_order (int, optional): Interpolation order to use when sampling pixels. Defaults to 1.
        oversample_factor (int, optional):
            Sample this number times more pixels in each dimension. This helps avoid aliasing but
            increases runtime. Defaults to 1.
        return_mask (bool, optional):
           Return a mask showing what pixels from the original image were
             sampled. Defaults to False.

    Returns:
        np.ndarray: The sampled perspective image
        Optional[np.ndarray]]: The mask of sampled locations, if requested
    """
    # Convert an equirectangular (360) image to a perspective view.
    H, W = equi_img.shape[:2]

    # Output image dimensions
    out_h, out_w = output_size

    # Sample a larger image and then downsample to reduce aliasing
    out_w = int(out_w * oversample_factor)
    out_h = int(out_h * oversample_factor)

    # FOV and angles in radians
    fov = np.deg2rad(fov_deg)

    # Compute the aspect ratio of the requested image dimensions
    aspect_ratio = out_h / out_w

    # Compute the distance from the center of the image to the edge in the x dimension
    x_dist = np.tan(fov / 2)
    y_dist = x_dist * aspect_ratio

    # Compute the size of each pixel to center the rays within the pixel rather than using the
    # top-left corner
    pixel_width = (2 * x_dist) / out_w

    # Create homogenous image coordinates for the output image
    x = np.arange(-x_dist + pixel_width / 2, x_dist, pixel_width)
    y = np.arange(-y_dist + pixel_width / 2, y_dist, pixel_width)

    # invert y for image coordinates to account for array indexing notation
    xv, yv = np.meshgrid(x, -y)

    # The z direction
    zv = np.ones_like(xv)

    pixel_directions = np.stack((xv, yv, zv), axis=-1)
    # normalize to unit
    pixel_directions /= np.linalg.norm(pixel_directions, axis=-1, keepdims=True)

    rotation_matrix = rotate_by_roll_pitch_yaw(roll_deg, pitch_deg, yaw_deg)

    # Rotate the pixel directions by the rotation matrix
    # The strange convention here is to deal with the fact that pixel_directions is (w, h, 3)
    # so this allows the dimensions to align for the matrix multiplication.
    pixel_directions = pixel_directions @ rotation_matrix.T

    # Convert 3D directions to spherical coordinates
    # horizontal angle
    horizontal = np.arctan2(pixel_directions[..., 0], pixel_directions[..., 2])
    # vertical angle. Clip to avoid floating point errors which extend beyond the valid domain
    # of arcsin
    altitude = np.arcsin(np.clip(pixel_directions[..., 1], -1.0, 1.0))

    # Map to equirectangular image coordinates
    i = (0.5 - altitude / np.pi) * H
    j = (horizontal / (2 * np.pi) + 0.5) * W
    # The range of these coordinates would be i in [0, H], j in [0, W]
    # To account for the wraparound effects if the sampling goes of the right edge, add the left row
    # of pixels to the right side.
    # TODO think more about sub-pixel considerations and how these coordinates interact with skimage
    # conventions on sampling and interpolation.
    equi_img = np.concatenate([equi_img, equi_img[:, 0:1, :]], axis=1)

    # Ensure that the coordinates are within the image. Note that the cropping in j is more generous
    # because of the wrap-around pixel.
    i = np.clip(i, 0, H - 1)
    j = np.clip(j, 0, W)

    # Stack the coordinates
    ij = np.stack((i, j), axis=0)

    # Sample pixels from the specified coordinates to obtain the perspective image
    sampled_perspective = flexible_inputs_warp(
        input_image=equi_img, inverse_map=ij, interpolation_order=warp_order
    )

    if oversample_factor > 1:
        sampled_perspective = downscale_local_mean(
            sampled_perspective, (oversample_factor, oversample_factor, 1)
        )

    # If only the resampled image is needed, return it
    if not return_mask:
        return sampled_perspective
    # Otherwise return a mask of the pixels which were sampled

    # Also save a mask of the pixels being sampled.
    mask = np.zeros((equi_img.shape[0], equi_img.shape[1]), dtype=bool)
    # Set pixels which were sampled to True. Out of bound errors are not possible because the values
    # have been clipped to the appropriate range.
    mask[np.round(i).astype(int), np.round(j).astype(int)] = True

    # In some cases the right edge might have been sampled, so report this as a sampled pixel on the
    # left edge
    mask[:, 0] = np.logical_or(mask[:, 0], mask[:, -1])
    # Remove the right edge pixels
    mask = mask[:, :-1]

    return sampled_perspective, mask



import argparse
import typing

import numpy as np
import pyproj

from geograypher.cameras import MetashapeCameraSet
from geograypher.constants import PATH_TYPE
from geograypher.meshes import TexturedPhotogrammetryMesh
from geograypher.utils.indexing import find_argmax_nonzero_value


def visualize(
    mesh_file: PATH_TYPE,
    mesh_CRS: typing.Union[pyproj.CRS, int],
    camera_file: PATH_TYPE,
    texture: typing.Optional[PATH_TYPE] = None,
    texture_column_name: typing.Optional[str] = None,
    IDs_to_labels: typing.Union[PATH_TYPE, dict, None] = None,
    downsample_target: float = 1.0,
    ROI: typing.Optional[PATH_TYPE] = None,
    ROI_buffer_meters: float = 0.0,
    convert_texture_to_max_class: bool = False,
):
    """
    Utility for visualizing meshes and associated camera sets. Note that this should be run in a
    graphical desktop.


    Args:
        mesh_file (PATH_TYPE):
            Path to the mesh file (often .ply)
        mesh_CRS (typing.Union[pyproj.CRS, int]):
            The vertex coordinates of the input mesh should be interpreteted in this coordinate
            references system to georeference them. Since meshes are not commonly used for
            geospatial tasks, there isn't a common standard for encoding this information in the mesh.
        camera_file (PATH_TYPE):
            Path to the MetaShape-exported .xml cameras file
        texture (typing.Union[PATH_TYPE, None]):
            See TexturedPhotogrammetryMesh.load_texture
        texture_column_name (typing.Union[str, None], optional):
            Column to use in vector file for texture information". Defaults to None.
        IDs_to_labels (typing.Union[PATH_TYPE, dict, None], optional):
            Mapping between the integer labels and string values for the classes or a path to a
            .json file representing the same. Defaults to None.
        downsample_target (float):
            Downsample the mesh to this fraction of vertices for increased performance but lower
            quality. Defaults to 1.
        ROI (PATH, optional):
            The region of interest to render labels for. Defaults to None.
        ROI_buffer_meters (typing.Optional[float]):
            The distance in meters to include around the ROI for the mesh. Defaults to 0.0.
        convert_texture_to_max_class (bool, optional):
            If the texture is a path to an .npy file encoding a (n_faces, n_classes) matrix of
            per-class weights, convert this representation into an (n_faces,) representation
            indicating the most common non-zero class. This is helpful for displaying the results
            of aggregated classification predictions. Defaults to False.
    """

    if convert_texture_to_max_class:
        # Load the texture
        texture = np.load(texture)
        # Convert to the argmax representation with nans for rows where all values were zero
        texture = find_argmax_nonzero_value(texture)

    # Load the camera set if provided
    if camera_file is not None:
        camera_set = MetashapeCameraSet(camera_file=camera_file, image_folder="")
    else:
        camera_set = None

    # If the camera set is provided and the ROI is not None, subset the cameras to that ROI
    if camera_set is not None and ROI is not None:
        camera_set = camera_set.get_subset_ROI(ROI, buffer_radius=ROI_buffer_meters)

    # Create the mesh
    mesh = TexturedPhotogrammetryMesh(
        mesh_file,
        input_CRS=mesh_CRS,
        ROI=ROI,
        ROI_buffer_meters=ROI_buffer_meters,
        downsample_target=downsample_target,
        texture=texture,
        texture_column_name=texture_column_name,
        IDs_to_labels=IDs_to_labels,
    )

    # Visualize the mesh and the camera set if provided
    # TODO optional visualization arguments like the colormap could be provided here
    mesh.vis(camera_set=camera_set)


def parse_args():
    """Parse and return arguements

    Returns:
        argparse.Namespace: Arguments
    """
    description = (
        "This script visualizes a mesh, optionally with added textures or a camera set. "
        + " All arguments are passed to geograypher.entrypoints.visualize.visualize "
        + "which has the following documentation:\n\n"
        + visualize.__doc__
    )
    # Ideally we'd include the defaults for each argument, but there is no help text so the
    # ArgumentDefaultsHelpFormatter formatter doesn't show them
    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser = argparse.ArgumentParser()
    parser.add_argument("--mesh-file", required=True)
    parser.add_argument("--mesh-CRS", required=True)
    parser.add_argument("--camera-file", required=True)
    parser.add_argument("--texture")
    parser.add_argument("--texture-column-name")
    parser.add_argument(
        "--IDs-to-labels",
        help="When using the CLI, path to a .json file must be provided, not a dict representation",
    )
    parser.add_argument("--downsample-target", type=float, default=1.0)
    parser.add_argument("--ROI")
    parser.add_argument("--ROI-buffer-meters", type=float, default=0.0)
    parser.add_argument("--convert-texture-to-max-class", action="store_true")

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = parse_args()
    visualize(**args.__dict__)



import logging

import numpy as np
import pandas as pd
from scipy.interpolate import griddata


def find_argmax_nonzero_value(
    array: np.ndarray, keepdims: bool = False, axis: int = 1
) -> np.array:
    """Find the argmax of an array, setting entires with zero sum or finite values to nan

    Args:
        array (np.ndarray): The input array
        keepdims (bool, optional): Should the dimensions be kept. Defaults to False.
        axis (int, optional): Which axis to perform the argmax along. Defaults to 1.

    Returns:
        np.array: The argmax, with nans for invalid or all-zero entries
    """
    # Find the column with the highest value per row
    argmax = np.argmax(array, axis=axis, keepdims=keepdims).astype(float)

    # Find rows with zero sum or any infinite values
    zero_sum_mask = np.sum(array, axis=axis) == 0
    infinite_mask = np.any(~np.isfinite(array), axis=axis)

    # Set these rows in the argmax to nan
    argmax[np.logical_or(zero_sum_mask, infinite_mask)] = np.nan

    return argmax


def ensure_float_labels(query_array, full_array=None) -> (np.ndarray, dict):
    # Standardizing the type
    if isinstance(query_array, pd.Series):
        query_array = query_array.to_numpy()
    else:
        query_array = np.array(query_array)

    # Check if the array is a string or some other object type that is not numeric
    if query_array.dtype in (str, np.dtype("O")):
        # Get unique values from full array if provided, or query array if not
        # Note that if the full array is not provided, the IDs will change based
        # on what set of the labeled data is indexed
        unique_values = np.unique(query_array if full_array is None else full_array)

        # Build an output array
        output_query_array = np.full_like(query_array, fill_value=np.nan, dtype=float)

        IDs_to_label = {}
        # Iterate through and set the values that match to the integer label
        for i, unique_value in enumerate(unique_values):
            output_query_array[query_array == unique_value] = i
            IDs_to_label[i] = unique_value

        return output_query_array, IDs_to_label

    # This is numeric data, but we want to check if it's representing something
    # categorical/ordinal

    # Determine which labels are finite so we can cast them to ints
    finite_labels = query_array[np.isfinite(query_array)]
    # See if all labels are approximately ints
    if np.allclose(finite_labels, finite_labels.astype(int)):
        # Try to remove any numerical issues by rounding to the ints
        unique_values = np.unique(
            np.round(query_array[np.isfinite(query_array)])
            if full_array is None
            else np.round(full_array[np.isfinite(full_array)])
        )

        IDs_to_label = {i: val for i, val in enumerate(unique_values)}
    else:
        # These are not discrete, so it doesn't make sense to represent them with IDs
        IDs_to_label = None

    # Cast to float, since that's expected
    output_query_array = query_array.astype(float)
    return output_query_array, IDs_to_label


def inverse_map_interpolation(
    ijmap: np.ndarray, downsample: int = 1, fill: int = -1
) -> np.ndarray:
    """
    Inverts the type of map that can be fed to skimage.transform.warp.

    The basic construction is the pixel position of these maps is the position
    in the destination image, and the value of the map is the position in the
    source image. Therefore if location [20, 30] has value [22.2, 28.4], it
    means that the destination image pixel [20, 30] will be sampled from the
    source image at pixel [22.2, 28.4] (the sampler can choose to snap to the
    closest integer value or interpolate nearby pixels).

    By inverting, we seek to reverse this map through interpolation. In the
    example above, in the output we would now have the location [22, 28]
    have the value [20, 30] (or slightly different because it might
    interpolate with nearby values).

    Arguments:
        ijmap ((2, H, W) numpy array): a map of the structure discussed above
        downsample (int): Inverting (particularly griddata) can be expensive
            for high-res image maps. You can downsample in integer steps to
            save computation, which will downsample both the (i, j) axes. Using
            downsample=2 will therefore interpolate on 1/4 of the pixels.
        fill (int): Values outside of the interpolation convex hull will take
            this value.

    Returns:
        (2, H, W) numpy array of the same shape as ijmap
    """

    # (row, col) grid of pixels coordinates
    H, W = ijmap.shape[1:]
    igrid, jgrid = np.meshgrid(np.arange(H), np.arange(W), indexing="ij")

    # Get an (N, 2) array of the grid coordinates
    grid_coords = np.stack([igrid.ravel(), jgrid.ravel()], axis=1)

    # Get (N, 2) arrays of the source and goal coordinates we have data for
    if downsample > 1:
        ds = slice(None, None, downsample)
        sample_y = np.stack([igrid[ds, ds].ravel(), jgrid[ds, ds].ravel()], axis=1)
        sample_x = np.stack(
            [ijmap[0][ds, ds].ravel(), ijmap[1][ds, ds].ravel()], axis=1
        )
    else:
        sample_y = grid_coords.copy()
        sample_x = np.stack([ijmap[0].ravel(), ijmap[1].ravel()], axis=1)

    # This is a little complicated, but griddata takes in
    # 1) The samples you have data for (x)
    # 2) The sample data (y)
    # 3) The new x at which you want to resample
    # In this case our sample x data is the mapped indices, the sample y data
    # is the grid that mapping came from, and the resample x is also the grid
    # because we are trying to invert.
    inv_i = griddata(
        sample_x, sample_y[:, 0], grid_coords, method="linear", fill_value=fill
    )
    inv_j = griddata(
        sample_x, sample_y[:, 1], grid_coords, method="linear", fill_value=fill
    )

    return np.stack([inv_i.reshape(H, W), inv_j.reshape(H, W)], axis=0)



import argparse
import json
from pathlib import Path

from geograypher.predictors.ortho_segmentor import write_chips


def parse_args():
    """Parse and return arguements

    Returns:
        argparse.Namespace: Arguments
    """

    description = (
        "This entrypoint is used for chipping raster data and optionally paired vector data. "
        + "All the arguments are passed to geograypher.entrypoints.write_chips "
        + "which has the following documentation:\n\n"
        + write_chips.__doc__
    )
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter, description=description
    )

    parser.add_argument("raster_file", type=Path)
    parser.add_argument("output_folder", type=Path)

    parser.add_argument("chip_size", type=int)
    parser.add_argument("chip_stride", type=int)

    parser.add_argument(
        "--label-vector-file", help="Path to vector label file.", type=Path
    )
    parser.add_argument("--label-column", type=str)
    parser.add_argument(
        "--label-remap",
        type=str,
        help="Provide this argument as a json-formatted string. "
        + 'For example \'{"name 1": 0, "name 2": 1}\'',
    )
    parser.add_argument("--write-empty-tiles", action="store_true")

    args = parser.parse_args()
    # Convert the json string representation to a dict
    args.label_remap = json.loads(args.label_remap)
    return args


if __name__ == "__main__":
    # Parse the arguments
    args = parse_args()
    # Pass them to the function
    write_chips(**args.__dict__)



"""
Render height masks made from a mesh and digital terrain model (DTM), saving renders
from each camera's perspective. There are two options:
- Threshold - Render a discrete mask with below ground, above ground, and invalid points.
    Determining which points are below vs. above ground is done by checking the height
    of mesh vertices above the DTM against a threshold.
- Raw - Render the mesh from each camera view, where the pixel value is the height of
    the mesh at that point above the STM. The output is an (M, N) numpy array instead
    of an image file.
"""

import argparse
import typing
from pathlib import Path

import numpy as np
import pyproj
import pyvista as pv
from matplotlib.pyplot import Normalize, cm

from geograypher.cameras.derived_cameras import MetashapeCameraSet
from geograypher.constants import PATH_TYPE
from geograypher.meshes.meshes import TexturedPhotogrammetryMesh
from geograypher.utils.files import ensure_folder
from geograypher.utils.visualization import show_segmentation_labels


def parse_args():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--image-folder",
        type=Path,
        required=True,
        help="Path to the folder containing images matching the camera XML data"
        " or a subset thereof. This script will iterate over available images"
        " in this folder.",
    )
    parser.add_argument(
        "--camera-file",
        type=Path,
        required=True,
        help="Path to the Metashape camera XML file with camera positions.",
    )
    parser.add_argument(
        "--mesh-file",
        type=Path,
        required=True,
        help="Path to the mesh file (e.g., .ply) that we will assess point height on.",
    )
    parser.add_argument(
        "--dtm-file",
        type=Path,
        required=True,
        help="Path to the digital terrain model (DTM) raster file (usually a tif).",
    )
    parser.add_argument(
        "--mesh-crs",
        type=int,
        required=True,
        help="The CRS to interpret the mesh in.",
    )
    parser.add_argument(
        "--original-image-folder",
        type=Path,
        required=False,
        help="If provided, this will be subtracted off the beginning of absolute image paths"
        " stored in the camera_file. See MetashapeCameraSet for details.",
    )
    parser.add_argument(
        "--output-folder",
        type=Path,
        required=True,
        help="Folder to save the rendered ground masks. Will be created if it doesn't exist",
    )
    parser.add_argument(
        "--output-mode",
        type=str,
        choices=["threshold", "raw"],
        default="raw",
        help="How to render the output: 'threshold' will render scenes with values"
        " 0=invalid, 1=below cutoff, 2=above cutoff; 'raw' will render scenes with the"
        " raw height values from the camera perspective.",
    )
    parser.add_argument(
        "--threshold-cutoff",
        type=float,
        default=1.0,
        help="Height threshold (same units as DTM → m) for ground/aboveground separation."
        " Only used if --output-mode is 'threshold'.",
    )
    parser.add_argument(
        "--vis-folder",
        type=Path,
        required=False,
        help="If provided, a textured mesh and evaluation images will be saved in this folder.",
    )
    parser.add_argument(
        "--vis-n-images",
        type=int,
        default=10,
        help="Number of eval images to save in the vis folder (only used if --vis-folder given).",
    )
    args = parser.parse_args()

    # Assert existence of all required Path arguments except output_folders
    assert (
        args.image_folder.exists()
    ), f"Image folder does not exist: {args.image_folder}"
    assert (
        args.camera_file.exists()
    ), f"Camera XML file does not exist: {args.camera_file}"
    assert args.dtm_file.exists(), f"DTM file does not exist: {args.dtm_file}"
    assert args.mesh_file.exists(), f"Mesh file does not exist: {args.mesh_file}"

    # Ensure output folder exists
    ensure_folder(args.output_folder)
    # Ensure vis folder exists if provided
    if args.vis_folder is not None:
        ensure_folder(args.vis_folder)

    return args


def render_height_masks(
    image_folder: PATH_TYPE,
    camera_file: PATH_TYPE,
    mesh_file: PATH_TYPE,
    dtm_file: PATH_TYPE,
    mesh_CRS: pyproj.CRS,
    original_image_folder: typing.Optional[PATH_TYPE],
    output_folder: PATH_TYPE,
    output_mode: str,
    threshold_cutoff: float,
    vis_folder: typing.Optional[PATH_TYPE],
    vis_n_images: int,
):
    """
    Render height masks made from a mesh and digital terrain model (DTM), saving renders
    geofrom each camera's perspective.

    Arguments:
        image_folder: PATH_TYPE, Path to the folder containing images matching the camera
            XML data or a subset thereof. This script will iterate over available images
            in this folder.
        camera_file: PATH_TYPE, Path to the Metashape camera XML file with camera positions.
        mesh_file: PATH_TYPE, Path to the mesh file (e.g., .ply) that we will assess point
            height on.
        dtm_file: PATH_TYPE, Path to the digital terrain model (DTM) raster file (usually a tif).
        mesh_CRS: pyproj.CRS, the CRS to interpret the mesh in.
        original_image_folder: typing.Optional[PATH_TYPE], If provided, this will be subtracted
            off the beginning of absolute image paths stored in the camera_file. See
            MetashapeCameraSet for details.
        output_folder: PATH_TYPE, Folder to save the rendered ground masks. Will be created
            if it doesn't exist
        output_mode: str, How to render the output: 'threshold' will render scenes with values
            0=invalid, 1=below cutoff, 2=above cutoff; 'raw' will render scenes with the
            raw height values from the camera perspective.
        threshold_cutoff: float, Height threshold (same units as DTM → m) for ground/aboveground
            separation. Only used if output_mode is 'threshold'.
        vis_folder: typing.Optional[PATH_TYPE], If provided, a textured mesh and evaluation
            images will be saved in this folder.
        vis_n_images: int, Number of eval images to save in the vis folder (only used if
            vis_folder given).

    Raises:
        ValueError: the output mode is an invalid mode
    """

    def load_mesh(texture=None):
        """Small helper function for something we repeat."""
        return TexturedPhotogrammetryMesh(
            mesh_file,
            input_CRS=mesh_CRS,
            texture=texture,
        )

    mesh = load_mesh()

    # Calculate the height of each mesh vertex above the detected ground (DTM).
    # Note that DTM files usually cover a smaller area spatially than the mesh,
    # and points outside the DTM ROI will have a height of NaN.
    height = mesh.get_height_above_ground(DTM_file=dtm_file)

    if output_mode == "threshold":
        # Sort the ground height into mesh texture, so that 0=nan, 1=ground, 2=aboveground
        texture = np.zeros(len(height), dtype=float)
        texture[np.isnan(height)] = 0
        texture[(~np.isnan(height)) & (height <= threshold_cutoff)] = 1
        texture[(~np.isnan(height)) & (height > threshold_cutoff)] = 2
        cast_to_uint8 = True
        label_suffix = ".tif"
        save_as_npy = False
    elif output_mode == "raw":
        # Just use the raw height values to retexture the mesh
        texture = height
        cast_to_uint8 = False
        label_suffix = ".npy"
        save_as_npy = True
    else:
        raise ValueError(f"Unknown mode: {output_mode}")

    # Reload the same mesh, but applying the height-labeled texture to the vertices
    height_mesh = load_mesh(texture=texture.reshape(-1, 1))

    # Load camera metadata
    camera_set = MetashapeCameraSet(
        camera_file,
        image_folder,
        original_image_folder=original_image_folder,
        validate_images=True,
    )
    extension = Path(camera_set.cameras[0].get_image_filename()).suffix

    if vis_folder is not None:
        # Save an evaluation mesh
        if output_mode == "threshold":
            # Note that we have to divide by 10 to get (0, 1, 2) to fit nicely into
            # the 0-1 range of tab10
            colored = cm.get_cmap("tab10")(texture.flatten() / 10)
        else:
            normalize = Normalize(vmin=np.nanmin(texture), vmax=np.nanmax(texture))
            colored = cm.get_cmap("viridis")(normalize(texture))
        vis_mesh = load_mesh(texture=(colored[:, :3] * 255).astype(np.uint8))
        vis_mesh.save_mesh(vis_folder / "height_mesh.ply", save_vert_texture=True)

    # For each camera, render the height-painted mesh onto that camera view
    height_mesh.save_renders(
        camera_set,
        output_folder=output_folder,
        save_native_resolution=True,
        cast_to_uint8=cast_to_uint8,
        save_as_npy=save_as_npy,
    )

    if vis_folder is not None:
        # Save evaluation images of the ground masks
        show_segmentation_labels(
            label_folder=output_folder,
            image_folder=image_folder,
            savefolder=vis_folder,
            num_show=vis_n_images,
            image_suffix=extension,
            label_suffix=label_suffix,
        )


if __name__ == "__main__":
    args = parse_args()
    render_height_masks(
        image_folder=args.image_folder,
        camera_file=args.camera_file,
        mesh_file=args.mesh_file,
        dtm_file=args.dtm_file,
        mesh_CRS=args.mesh_crs,
        original_image_folder=args.original_image_folder,
        output_folder=args.output_folder,
        output_mode=args.output_mode,
        threshold_cutoff=args.threshold_cutoff,
        vis_folder=args.vis_folder,
        vis_n_images=args.vis_n_images,
    )
